# Natural Language Toolkit: Simple Tokenizers
#
# Copyright (C) 2001-2018 NLTK Project
# Author: Edward Loper <edloper@gmail.com>
#         Steven Bird <stevenbird1@gmail.com>
# URL: <http://nltk.sourceforge.net>
# For license information, see LICENSE.TXT

r"""
Simple Tokenizers

These tokenizers divide strings into substrings using the string
``split()`` method.
When tokenizing using a particular delimiter string, use
the string ``split()`` method directly, as this is more efficient.

The simple tokenizers are *not* available as separate functions;
instead, you should just use the string ``split()`` method directly:

    >>> s = "Good muffins cost $3.88\nin New York.  Please buy me\ntwo of them.\n\nThanks."
    >>> s.split()
    ['Good', 'muffins', 'cost', '$3.88', 'in', 'New', 'York.',
    'Please', 'buy', 'me', 'two', 'of', 'them.', 'Thanks.']
    >>> s.split(' ')
    ['Good', 'muffins', 'cost', '$3.88\nin', 'New', 'York.', '',
    'Please', 'buy', 'me\ntwo', 'of', 'them.\n\nThanks.']
    >>> s.split('\n')
    ['Good muffins cost $3.88', 'in New York.  Please buy me',
    'two of them.', '', 'Thanks.']

The simple tokenizers are mainly useful because they follow the
standard ``TokenizerI`` interface, and so can be used with any code
that expects a tokenizer.  For example, these tokenizers can be used
to specify the tokenization conventions when building a `CorpusReader`.

"""
from __future__ import unicode_literals
from nltk.tokenize.api import TokenizerI, StringTokenizer
from nltk.tokenize.util import string_span_tokenize, regexp_span_tokenize


class SpaceTokenizer(StringTokenizer):
    r"""Tokenize a string using the space character as a delimiter,
    which is the same as ``s.split(' ')``.

        >>> from nltk.tokenize import SpaceTokenizer
        >>> s = "Good muffins cost $3.88\nin New York.  Please buy me\ntwo of them.\n\nThanks."
        >>> SpaceTokenizer().tokenize(s)
        ['Good', 'muffins', 'cost', '$3.88\nin', 'New', 'York.', '',
        'Please', 'buy', 'me\ntwo', 'of', 'them.\n\nThanks.']
    """

    _string = ' '


class TabTokenizer(StringTokenizer):
    r"""Tokenize a string use the tab character as a delimiter,
    the same as ``s.split('\t')``.

        >>> from nltk.tokenize import TabTokenizer
        >>> TabTokenizer().tokenize('a\tb c\n\t d')
        ['a', 'b c\n', ' d']
    """

    _string = '\t'


class CharTokenizer(StringTokenizer):
    """Tokenize a string into individual characters.  If this functionality
    is ever required directly, use ``for char in string``.
    """

    def tokenize(self, s):
        return list(s)

    def span_tokenize(self, s):
        for i, j in enumerate(range(1, len(s) + 1)):
            yield i, j


class LineTokenizer(TokenizerI):
    r"""Tokenize a string into its lines, optionally discarding blank lines.
    This is similar to ``s.split('\n')``.

        >>> from nltk.tokenize import LineTokenizer
        >>> s = "Good muffins cost $3.88\nin New York.  Please buy me\ntwo of them.\n\nThanks."
        >>> LineTokenizer(blanklines='keep').tokenize(s)
        ['Good muffins cost $3.88', 'in New York.  Please buy me',
        'two of them.', '', 'Thanks.']
        >>> # same as [l for l in s.split('\n') if l.strip()]:
        >>> LineTokenizer(blanklines='discard').tokenize(s)
        ['Good muffins cost $3.88', 'in New York.  Please buy me',
        'two of them.', 'Thanks.']

    :param blanklines: Indicates how blank lines should be handled.  Valid values are:

        - ``discard``: strip blank lines out of the token list before returning it.
           A line is considered blank if it contains only whitespace characters.
        - ``keep``: leave all blank lines in the token list.
        - ``discard-eof``: if the string ends with a newline, then do not generate
           a corresponding token ``''`` after that newline.
    """

    def __init__(self, blanklines='discard'):
        valid_blanklines = ('discard', 'keep', 'discard-eof')
        if blanklines not in valid_blanklines:
            raise ValueError('Blank lines must be one of: %s' %
                             ' '.join(valid_blanklines))

        self._blanklines = blanklines

    def tokenize(self, s):
        lines = s.splitlines()
        # If requested, strip off blank lines.
        if self._blanklines == 'discard':
            lines = [l for l in lines if l.rstrip()]
        elif self._blanklines == 'discard-eof':
            if lines and not lines[-1].strip():
                lines.pop()
        return lines

    # discard-eof not implemented
    def span_tokenize(self, s):
        if self._blanklines == 'keep':
            for span in string_span_tokenize(s, r'\n'):
                yield span
        else:
            for span in regexp_span_tokenize(s, r'\n(\s+\n)*'):
                yield span


######################################################################
# { Tokenization Functions
######################################################################
# XXX: it is stated in module docs that there is no function versions

def line_tokenize(text, blanklines='discard'):
    return LineTokenizer(blanklines).tokenize(text)



# -*- coding: utf-8 -*-
"""
UDHR corpus reader. It mostly deals with encodings.
"""
from __future__ import absolute_import, unicode_literals

from nltk.corpus.reader.util import find_corpus_fileids
from nltk.corpus.reader.plaintext import PlaintextCorpusReader

class UdhrCorpusReader(PlaintextCorpusReader):

    ENCODINGS = [
        ('.*-Latin1$', 'latin-1'),
        ('.*-Hebrew$', 'hebrew'),
        ('.*-Arabic$', 'cp1256'),
        ('Czech_Cesky-UTF8', 'cp1250'), # yeah
        ('.*-Cyrillic$', 'cyrillic'),
        ('.*-SJIS$', 'SJIS'),
        ('.*-GB2312$', 'GB2312'),
        ('.*-Latin2$', 'ISO-8859-2'),
        ('.*-Greek$', 'greek'),
        ('.*-UTF8$', 'utf-8'),

        ('Hungarian_Magyar-Unicode', 'utf-16-le'),
        ('Amahuaca', 'latin1'),
        ('Turkish_Turkce-Turkish', 'latin5'),
        ('Lithuanian_Lietuviskai-Baltic', 'latin4'),
        ('Japanese_Nihongo-EUC', 'EUC-JP'),
        ('Japanese_Nihongo-JIS', 'iso2022_jp'),
        ('Chinese_Mandarin-HZ', 'hz'),
        ('Abkhaz\-Cyrillic\+Abkh', 'cp1251'),
    ]

    SKIP = set([
        # The following files are not fully decodable because they
        # were truncated at wrong bytes:
        'Burmese_Myanmar-UTF8',
        'Japanese_Nihongo-JIS',
        'Chinese_Mandarin-HZ',
        'Chinese_Mandarin-UTF8',
        'Gujarati-UTF8',
        'Hungarian_Magyar-Unicode',
        'Lao-UTF8',
        'Magahi-UTF8',
        'Marathi-UTF8',
        'Tamil-UTF8',

        # Unfortunately, encodings required for reading
        # the following files are not supported by Python:
        'Vietnamese-VPS',
        'Vietnamese-VIQR',
        'Vietnamese-TCVN',
        'Magahi-Agra',
        'Bhojpuri-Agra',
        'Esperanto-T61', # latin3 raises an exception

        # The following files are encoded for specific fonts:
        'Burmese_Myanmar-WinResearcher',
        'Armenian-DallakHelv',
        'Tigrinya_Tigrigna-VG2Main',
        'Amharic-Afenegus6..60375', # ?
        'Navaho_Dine-Navajo-Navaho-font',

        # What are these?
        'Azeri_Azerbaijani_Cyrillic-Az.Times.Cyr.Normal0117',
        'Azeri_Azerbaijani_Latin-Az.Times.Lat0117',

        # The following files are unintended:
        'Czech-Latin2-err',
        'Russian_Russky-UTF8~',
    ])


    def __init__(self, root='udhr'):
        fileids = find_corpus_fileids(root, r'(?!README|\.).*')
        super(UdhrCorpusReader, self).__init__(
            root,
            [fileid for fileid in fileids if fileid not in self.SKIP],
            encoding=self.ENCODINGS
        )



# Natural Language Toolkit: Rude Chatbot
#
# Copyright (C) 2001-2018 NLTK Project
# Author: Peter Spiller <pspiller@csse.unimelb.edu.au>
# URL: <http://nltk.org/>
# For license information, see LICENSE.TXT
from __future__ import print_function

from nltk.chat.util import Chat, reflections

pairs = (
    (r'We (.*)',
        ("What do you mean, 'we'?",
        "Don't include me in that!",
        "I wouldn't be so sure about that.")),

    (r'You should (.*)',
        ("Don't tell me what to do, buddy.",
        "Really? I should, should I?")),

    (r'You\'re(.*)',
        ("More like YOU'RE %1!",
        "Hah! Look who's talking.",
        "Come over here and tell me I'm %1.")),

    (r'You are(.*)',
        ("More like YOU'RE %1!",
        "Hah! Look who's talking.",
        "Come over here and tell me I'm %1.")),

    (r'I can\'t(.*)',
        ("You do sound like the type who can't %1.",
        "Hear that splashing sound? That's my heart bleeding for you.",
        "Tell somebody who might actually care.")),

    (r'I think (.*)',
        ("I wouldn't think too hard if I were you.",
        "You actually think? I'd never have guessed...")),

    (r'I (.*)',
        ("I'm getting a bit tired of hearing about you.",
        "How about we talk about me instead?",
        "Me, me, me... Frankly, I don't care.")),

    (r'How (.*)',
        ("How do you think?",
        "Take a wild guess.",
        "I'm not even going to dignify that with an answer.")),

    (r'What (.*)',
        ("Do I look like an encyclopedia?",
        "Figure it out yourself.")),

    (r'Why (.*)',
        ("Why not?",
        "That's so obvious I thought even you'd have already figured it out.")),

    (r'(.*)shut up(.*)',
        ("Make me.",
        "Getting angry at a feeble NLP assignment? Somebody's losing it.",
        "Say that again, I dare you.")),

    (r'Shut up(.*)',
        ("Make me.",
        "Getting angry at a feeble NLP assignment? Somebody's losing it.",
        "Say that again, I dare you.")),

    (r'Hello(.*)',
        ("Oh good, somebody else to talk to. Joy.",
        "'Hello'? How original...")),

    (r'(.*)',
        ("I'm getting bored here. Become more interesting.",
        "Either become more thrilling or get lost, buddy.",
        "Change the subject before I die of fatal boredom."))
)

rude_chatbot = Chat(pairs, reflections)

def rude_chat():
    print("Talk to the program by typing in plain English, using normal upper-")
    print('and lower-case letters and punctuation.  Enter "quit" when done.')
    print('='*72)
    print("I suppose I should say hello.")

    rude_chatbot.converse()

def demo():
    rude_chat()

if __name__ == "__main__":
    demo()



# Natural Language Toolkit: Tagger Interface
#
# Copyright (C) 2001-2018 NLTK Project
# Author: Edward Loper <edloper@gmail.com>
#         Steven Bird <stevenbird1@gmail.com> (minor additions)
# URL: <http://nltk.org/>
# For license information, see LICENSE.TXT

"""
Interface for tagging each token in a sentence with supplementary
information, such as its part of speech.
"""
from abc import ABCMeta, abstractmethod
from six import add_metaclass
from itertools import chain

from nltk.internals import overridden
from nltk.metrics import accuracy

from nltk.tag.util import untag


@add_metaclass(ABCMeta)
class TaggerI(object):
    """
    A processing interface for assigning a tag to each token in a list.
    Tags are case sensitive strings that identify some property of each
    token, such as its part of speech or its sense.

    Some taggers require specific types for their tokens.  This is
    generally indicated by the use of a sub-interface to ``TaggerI``.
    For example, featureset taggers, which are subclassed from
    ``FeaturesetTagger``, require that each token be a ``featureset``.

    Subclasses must define:
      - either ``tag()`` or ``tag_sents()`` (or both)
    """
    @abstractmethod
    def tag(self, tokens):
        """
        Determine the most appropriate tag sequence for the given
        token sequence, and return a corresponding list of tagged
        tokens.  A tagged token is encoded as a tuple ``(token, tag)``.

        :rtype: list(tuple(str, str))
        """
        if overridden(self.tag_sents):
            return self.tag_sents([tokens])[0]

    def tag_sents(self, sentences):
        """
        Apply ``self.tag()`` to each element of *sentences*.  I.e.:

            return [self.tag(sent) for sent in sentences]
        """
        return [self.tag(sent) for sent in sentences]

    def evaluate(self, gold):
        """
        Score the accuracy of the tagger against the gold standard.
        Strip the tags from the gold standard text, retag it using
        the tagger, then compute the accuracy score.

        :type gold: list(list(tuple(str, str)))
        :param gold: The list of tagged sentences to score the tagger on.
        :rtype: float
        """

        tagged_sents = self.tag_sents(untag(sent) for sent in gold)
        gold_tokens = list(chain(*gold))
        test_tokens = list(chain(*tagged_sents))
        return accuracy(gold_tokens, test_tokens)

    def _check_params(self, train, model):
        if (train and model) or (not train and not model):
            raise ValueError(
                    'Must specify either training data or trained model.')


class FeaturesetTaggerI(TaggerI):
    """
    A tagger that requires tokens to be ``featuresets``.  A featureset
    is a dictionary that maps from feature names to feature
    values.  See ``nltk.classify`` for more information about features
    and featuresets.
    """



# -*- coding: utf-8 -*-

"""
Mock test for Stanford CoreNLP wrappers.
"""

import sys
from itertools import chain
from unittest import TestCase, SkipTest

try:
    from unittest.mock import MagicMock
except ImportError:
    raise SkipTest('unittest.mock no supported in Python2')
from nltk.tree import Tree
from nltk.parse import corenlp


class TestTokenizerAPI(TestCase):

    def test_tokenize(self):
        corenlp_tokenizer = corenlp.CoreNLPParser()

        api_return_value = {
            u'sentences': [{u'index': 0,
                            u'tokens': [{u'after': u' ',
                                         u'before': u'',
                                         u'characterOffsetBegin': 0,
                                         u'characterOffsetEnd': 4,
                                         u'index': 1,
                                         u'originalText': u'Good',
                                         u'word': u'Good'},
                                        {u'after': u' ',
                                         u'before': u' ',
                                         u'characterOffsetBegin': 5,
                                         u'characterOffsetEnd': 12,
                                         u'index': 2,
                                         u'originalText': u'muffins',
                                         u'word': u'muffins'},
                                        {u'after': u' ',
                                         u'before': u' ',
                                         u'characterOffsetBegin': 13,
                                         u'characterOffsetEnd': 17,
                                         u'index': 3,
                                         u'originalText': u'cost',
                                         u'word': u'cost'},
                                        {u'after': u'',
                                         u'before': u' ',
                                         u'characterOffsetBegin': 18,
                                         u'characterOffsetEnd': 19,
                                         u'index': 4,
                                         u'originalText': u'$',
                                         u'word': u'$'},
                                        {u'after': u'\n',
                                         u'before': u'',
                                         u'characterOffsetBegin': 19,
                                         u'characterOffsetEnd': 23,
                                         u'index': 5,
                                         u'originalText': u'3.88',
                                         u'word': u'3.88'},
                                        {u'after': u' ',
                                         u'before': u'\n',
                                         u'characterOffsetBegin': 24,
                                         u'characterOffsetEnd': 26,
                                         u'index': 6,
                                         u'originalText': u'in',
                                         u'word': u'in'},
                                        {u'after': u' ',
                                         u'before': u' ',
                                         u'characterOffsetBegin': 27,
                                         u'characterOffsetEnd': 30,
                                         u'index': 7,
                                         u'originalText': u'New',
                                         u'word': u'New'},
                                        {u'after': u'',
                                         u'before': u' ',
                                         u'characterOffsetBegin': 31,
                                         u'characterOffsetEnd': 35,
                                         u'index': 8,
                                         u'originalText': u'York',
                                         u'word': u'York'},
                                        {u'after': u'  ',
                                         u'before': u'',
                                         u'characterOffsetBegin': 35,
                                         u'characterOffsetEnd': 36,
                                         u'index': 9,
                                         u'originalText': u'.',
                                         u'word': u'.'}]},
                           {u'index': 1,
                            u'tokens': [{u'after': u' ',
                                         u'before': u'  ',
                                         u'characterOffsetBegin': 38,
                                         u'characterOffsetEnd': 44,
                                         u'index': 1,
                                         u'originalText': u'Please',
                                         u'word': u'Please'},
                                        {u'after': u' ',
                                         u'before': u' ',
                                         u'characterOffsetBegin': 45,
                                         u'characterOffsetEnd': 48,
                                         u'index': 2,
                                         u'originalText': u'buy',
                                         u'word': u'buy'},
                                        {u'after': u'\n',
                                         u'before': u' ',
                                         u'characterOffsetBegin': 49,
                                         u'characterOffsetEnd': 51,
                                         u'index': 3,
                                         u'originalText': u'me',
                                         u'word': u'me'},
                                        {u'after': u' ',
                                         u'before': u'\n',
                                         u'characterOffsetBegin': 52,
                                         u'characterOffsetEnd': 55,
                                         u'index': 4,
                                         u'originalText': u'two',
                                         u'word': u'two'},
                                        {u'after': u' ',
                                         u'before': u' ',
                                         u'characterOffsetBegin': 56,
                                         u'characterOffsetEnd': 58,
                                         u'index': 5,
                                         u'originalText': u'of',
                                         u'word': u'of'},
                                        {u'after': u'',
                                         u'before': u' ',
                                         u'characterOffsetBegin': 59,
                                         u'characterOffsetEnd': 63,
                                         u'index': 6,
                                         u'originalText': u'them',
                                         u'word': u'them'},
                                        {u'after': u'\n',
                                         u'before': u'',
                                         u'characterOffsetBegin': 63,
                                         u'characterOffsetEnd': 64,
                                         u'index': 7,
                                         u'originalText': u'.',
                                         u'word': u'.'}]},
                           {u'index': 2,
                            u'tokens': [{u'after': u'',
                                         u'before': u'\n',
                                         u'characterOffsetBegin': 65,
                                         u'characterOffsetEnd': 71,
                                         u'index': 1,
                                         u'originalText': u'Thanks',
                                         u'word': u'Thanks'},
                                        {u'after': u'',
                                         u'before': u'',
                                         u'characterOffsetBegin': 71,
                                         u'characterOffsetEnd': 72,
                                         u'index': 2,
                                         u'originalText': u'.',
                                         u'word': u'.'}]}]
        }
        corenlp_tokenizer.api_call = MagicMock(return_value=api_return_value)

        input_string = "Good muffins cost $3.88\nin New York.  Please buy me\ntwo of them.\nThanks."

        expected_output = [u'Good', u'muffins', u'cost', u'$', u'3.88', u'in',
                           u'New', u'York', u'.', u'Please', u'buy', u'me',
                           u'two', u'of', u'them', u'.', u'Thanks', u'.']

        tokenized_output = list(corenlp_tokenizer.tokenize(input_string))

        corenlp_tokenizer.api_call.assert_called_once_with(
            'Good muffins cost $3.88\nin New York.  Please buy me\ntwo of them.\nThanks.', properties={'annotators': 'tokenize,ssplit'})
        self.assertEqual(expected_output, tokenized_output)


class TestTaggerAPI(TestCase):

    def test_pos_tagger(self):
        corenlp_tagger = corenlp.CoreNLPParser(tagtype='pos')

        api_return_value = {
            u'sentences': [{u'basicDependencies': [{u'dep': u'ROOT',
                                                    u'dependent': 1,
                                                    u'dependentGloss': u'What',
                                                    u'governor': 0,
                                                    u'governorGloss': u'ROOT'},
                                                   {u'dep': u'cop',
                                                    u'dependent': 2,
                                                    u'dependentGloss': u'is',
                                                    u'governor': 1,
                                                    u'governorGloss': u'What'},
                                                   {u'dep': u'det',
                                                    u'dependent': 3,
                                                    u'dependentGloss': u'the',
                                                    u'governor': 4,
                                                    u'governorGloss': u'airspeed'},
                                                   {u'dep': u'nsubj',
                                                    u'dependent': 4,
                                                    u'dependentGloss': u'airspeed',
                                                    u'governor': 1,
                                                    u'governorGloss': u'What'},
                                                   {u'dep': u'case',
                                                    u'dependent': 5,
                                                    u'dependentGloss': u'of',
                                                    u'governor': 8,
                                                    u'governorGloss': u'swallow'},
                                                   {u'dep': u'det',
                                                    u'dependent': 6,
                                                    u'dependentGloss': u'an',
                                                    u'governor': 8,
                                                    u'governorGloss': u'swallow'},
                                                   {u'dep': u'compound',
                                                    u'dependent': 7,
                                                    u'dependentGloss': u'unladen',
                                                    u'governor': 8,
                                                    u'governorGloss': u'swallow'},
                                                   {u'dep': u'nmod',
                                                    u'dependent': 8,
                                                    u'dependentGloss': u'swallow',
                                                    u'governor': 4,
                                                    u'governorGloss': u'airspeed'},
                                                   {u'dep': u'punct',
                                                    u'dependent': 9,
                                                    u'dependentGloss': u'?',
                                                    u'governor': 1,
                                                    u'governorGloss': u'What'}],
                            u'enhancedDependencies': [{u'dep': u'ROOT',
                                                       u'dependent': 1,
                                                       u'dependentGloss': u'What',
                                                       u'governor': 0,
                                                       u'governorGloss': u'ROOT'},
                                                      {u'dep': u'cop',
                                                       u'dependent': 2,
                                                       u'dependentGloss': u'is',
                                                       u'governor': 1,
                                                       u'governorGloss': u'What'},
                                                      {u'dep': u'det',
                                                       u'dependent': 3,
                                                       u'dependentGloss': u'the',
                                                       u'governor': 4,
                                                       u'governorGloss': u'airspeed'},
                                                      {u'dep': u'nsubj',
                                                       u'dependent': 4,
                                                       u'dependentGloss': u'airspeed',
                                                       u'governor': 1,
                                                       u'governorGloss': u'What'},
                                                      {u'dep': u'case',
                                                       u'dependent': 5,
                                                       u'dependentGloss': u'of',
                                                       u'governor': 8,
                                                       u'governorGloss': u'swallow'},
                                                      {u'dep': u'det',
                                                       u'dependent': 6,
                                                       u'dependentGloss': u'an',
                                                       u'governor': 8,
                                                       u'governorGloss': u'swallow'},
                                                      {u'dep': u'compound',
                                                       u'dependent': 7,
                                                       u'dependentGloss': u'unladen',
                                                       u'governor': 8,
                                                       u'governorGloss': u'swallow'},
                                                      {u'dep': u'nmod:of',
                                                       u'dependent': 8,
                                                       u'dependentGloss': u'swallow',
                                                       u'governor': 4,
                                                       u'governorGloss': u'airspeed'},
                                                      {u'dep': u'punct',
                                                       u'dependent': 9,
                                                       u'dependentGloss': u'?',
                                                       u'governor': 1,
                                                       u'governorGloss': u'What'}],
                            u'enhancedPlusPlusDependencies': [{u'dep': u'ROOT',
                                                               u'dependent': 1,
                                                               u'dependentGloss': u'What',
                                                               u'governor': 0,
                                                               u'governorGloss': u'ROOT'},
                                                              {u'dep': u'cop',
                                                               u'dependent': 2,
                                                               u'dependentGloss': u'is',
                                                               u'governor': 1,
                                                               u'governorGloss': u'What'},
                                                              {u'dep': u'det',
                                                               u'dependent': 3,
                                                               u'dependentGloss': u'the',
                                                               u'governor': 4,
                                                               u'governorGloss': u'airspeed'},
                                                              {u'dep': u'nsubj',
                                                               u'dependent': 4,
                                                               u'dependentGloss': u'airspeed',
                                                               u'governor': 1,
                                                               u'governorGloss': u'What'},
                                                              {u'dep': u'case',
                                                               u'dependent': 5,
                                                               u'dependentGloss': u'of',
                                                               u'governor': 8,
                                                               u'governorGloss': u'swallow'},
                                                              {u'dep': u'det',
                                                               u'dependent': 6,
                                                               u'dependentGloss': u'an',
                                                               u'governor': 8,
                                                               u'governorGloss': u'swallow'},
                                                              {u'dep': u'compound',
                                                               u'dependent': 7,
                                                               u'dependentGloss': u'unladen',
                                                               u'governor': 8,
                                                               u'governorGloss': u'swallow'},
                                                              {u'dep': u'nmod:of',
                                                               u'dependent': 8,
                                                               u'dependentGloss': u'swallow',
                                                               u'governor': 4,
                                                               u'governorGloss': u'airspeed'},
                                                              {u'dep': u'punct',
                                                               u'dependent': 9,
                                                               u'dependentGloss': u'?',
                                                               u'governor': 1,
                                                               u'governorGloss': u'What'}],
                            u'index': 0,
                            u'parse': u'(ROOT\n  (SBARQ\n    (WHNP (WP What))\n    (SQ (VBZ is)\n      (NP\n        (NP (DT the) (NN airspeed))\n        (PP (IN of)\n          (NP (DT an) (NN unladen) (NN swallow)))))\n    (. ?)))',
                            u'tokens': [{u'after': u' ',
                                         u'before': u'',
                                         u'characterOffsetBegin': 0,
                                         u'characterOffsetEnd': 4,
                                         u'index': 1,
                                         u'lemma': u'what',
                                         u'originalText': u'What',
                                         u'pos': u'WP',
                                         u'word': u'What'},
                                        {u'after': u' ',
                                         u'before': u' ',
                                         u'characterOffsetBegin': 5,
                                         u'characterOffsetEnd': 7,
                                         u'index': 2,
                                         u'lemma': u'be',
                                         u'originalText': u'is',
                                         u'pos': u'VBZ',
                                         u'word': u'is'},
                                        {u'after': u' ',
                                         u'before': u' ',
                                         u'characterOffsetBegin': 8,
                                         u'characterOffsetEnd': 11,
                                         u'index': 3,
                                         u'lemma': u'the',
                                         u'originalText': u'the',
                                         u'pos': u'DT',
                                         u'word': u'the'},
                                        {u'after': u' ',
                                         u'before': u' ',
                                         u'characterOffsetBegin': 12,
                                         u'characterOffsetEnd': 20,
                                         u'index': 4,
                                         u'lemma': u'airspeed',
                                         u'originalText': u'airspeed',
                                         u'pos': u'NN',
                                         u'word': u'airspeed'},
                                        {u'after': u' ',
                                         u'before': u' ',
                                         u'characterOffsetBegin': 21,
                                         u'characterOffsetEnd': 23,
                                         u'index': 5,
                                         u'lemma': u'of',
                                         u'originalText': u'of',
                                         u'pos': u'IN',
                                         u'word': u'of'},
                                        {u'after': u' ',
                                         u'before': u' ',
                                         u'characterOffsetBegin': 24,
                                         u'characterOffsetEnd': 26,
                                         u'index': 6,
                                         u'lemma': u'a',
                                         u'originalText': u'an',
                                         u'pos': u'DT',
                                         u'word': u'an'},
                                        {u'after': u' ',
                                         u'before': u' ',
                                         u'characterOffsetBegin': 27,
                                         u'characterOffsetEnd': 34,
                                         u'index': 7,
                                         u'lemma': u'unladen',
                                         u'originalText': u'unladen',
                                         u'pos': u'JJ',
                                         u'word': u'unladen'},
                                        {u'after': u' ',
                                         u'before': u' ',
                                         u'characterOffsetBegin': 35,
                                         u'characterOffsetEnd': 42,
                                         u'index': 8,
                                         u'lemma': u'swallow',
                                         u'originalText': u'swallow',
                                         u'pos': u'VB',
                                         u'word': u'swallow'},
                                        {u'after': u'',
                                         u'before': u' ',
                                         u'characterOffsetBegin': 43,
                                         u'characterOffsetEnd': 44,
                                         u'index': 9,
                                         u'lemma': u'?',
                                         u'originalText': u'?',
                                         u'pos': u'.',
                                         u'word': u'?'}]}]
        }
        corenlp_tagger.api_call = MagicMock(return_value=api_return_value)

        input_tokens = 'What is the airspeed of an unladen swallow ?'.split()
        expected_output = [('What', 'WP'), ('is', 'VBZ'), ('the', 'DT'),
                           ('airspeed', 'NN'), ('of', 'IN'), ('an', 'DT'),
                           ('unladen', 'JJ'), ('swallow', 'VB'), ('?', '.')]
        tagged_output = corenlp_tagger.tag(input_tokens)

        corenlp_tagger.api_call.assert_called_once_with('What is the airspeed of an unladen swallow ?', properties={'ssplit.isOneSentence': 'true',
                                                                                                                    'annotators': 'tokenize,ssplit,pos'})
        self.assertEqual(expected_output, tagged_output)

    def test_ner_tagger(self):
        corenlp_tagger = corenlp.CoreNLPParser(tagtype='ner')

        api_return_value = {
            'sentences': [{'index': 0,
                           'tokens': [{'after': ' ',
                                       'before': '',
                                       'characterOffsetBegin': 0,
                                       'characterOffsetEnd': 4,
                                       'index': 1,
                                       'lemma': 'Rami',
                                       'ner': 'PERSON',
                                       'originalText': 'Rami',
                                                       'pos': 'NNP',
                                                       'word': 'Rami'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 5,
                                       'characterOffsetEnd': 8,
                                       'index': 2,
                                       'lemma': 'Eid',
                                       'ner': 'PERSON',
                                       'originalText': 'Eid',
                                                       'pos': 'NNP',
                                                       'word': 'Eid'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 9,
                                       'characterOffsetEnd': 11,
                                       'index': 3,
                                       'lemma': 'be',
                                       'ner': 'O',
                                       'originalText': 'is',
                                                       'pos': 'VBZ',
                                                       'word': 'is'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 12,
                                       'characterOffsetEnd': 20,
                                       'index': 4,
                                       'lemma': 'study',
                                       'ner': 'O',
                                       'originalText': 'studying',
                                                       'pos': 'VBG',
                                                       'word': 'studying'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 21,
                                       'characterOffsetEnd': 23,
                                       'index': 5,
                                       'lemma': 'at',
                                       'ner': 'O',
                                       'originalText': 'at',
                                                       'pos': 'IN',
                                                       'word': 'at'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 24,
                                       'characterOffsetEnd': 29,
                                       'index': 6,
                                       'lemma': 'Stony',
                                       'ner': 'ORGANIZATION',
                                       'originalText': 'Stony',
                                                       'pos': 'NNP',
                                                       'word': 'Stony'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 30,
                                       'characterOffsetEnd': 35,
                                       'index': 7,
                                       'lemma': 'Brook',
                                       'ner': 'ORGANIZATION',
                                       'originalText': 'Brook',
                                                       'pos': 'NNP',
                                                       'word': 'Brook'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 36,
                                       'characterOffsetEnd': 46,
                                       'index': 8,
                                       'lemma': 'University',
                                       'ner': 'ORGANIZATION',
                                       'originalText': 'University',
                                       'pos': 'NNP',
                                                       'word': 'University'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 47,
                                       'characterOffsetEnd': 49,
                                       'index': 9,
                                       'lemma': 'in',
                                       'ner': 'O',
                                       'originalText': 'in',
                                       'pos': 'IN',
                                       'word': 'in'},
                                      {'after': '',
                                       'before': ' ',
                                       'characterOffsetBegin': 50,
                                       'characterOffsetEnd': 52,
                                       'index': 10,
                                       'lemma': 'NY',
                                       'ner': 'O',
                                       'originalText': 'NY',
                                       'pos': 'NNP',
                                       'word': 'NY'}]}]}

        corenlp_tagger.api_call = MagicMock(return_value=api_return_value)

        input_tokens = 'Rami Eid is studying at Stony Brook University in NY'.split()
        expected_output = [('Rami', 'PERSON'), ('Eid', 'PERSON'), ('is', 'O'), ('studying', 'O'), ('at', 'O'), ('Stony', 'ORGANIZATION'),
                           ('Brook', 'ORGANIZATION'), ('University', 'ORGANIZATION'), ('in', 'O'), ('NY', 'O')]
        tagged_output = corenlp_tagger.tag(input_tokens)

        corenlp_tagger.api_call.assert_called_once_with('Rami Eid is studying at Stony Brook University in NY', properties={'ssplit.isOneSentence': 'true',
                                                                                                                            'annotators': 'tokenize,ssplit,ner'})
        self.assertEqual(expected_output, tagged_output)

    def test_unexpected_tagtype(self):
        with self.assertRaises(ValueError):
            corenlp_tagger = corenlp.CoreNLPParser(tagtype='test')

class TestParserAPI(TestCase):

    def test_parse(self):
        corenlp_parser = corenlp.CoreNLPParser()

        api_return_value = {
            'sentences': [{'basicDependencies': [{'dep': 'ROOT',
                                                  'dependent': 4,
                                                  'dependentGloss': 'fox',
                                                  'governor': 0,
                                                  'governorGloss': 'ROOT'},
                                                 {'dep': 'det',
                                                  'dependent': 1,
                                                  'dependentGloss': 'The',
                                                  'governor': 4,
                                                  'governorGloss': 'fox'},
                                                 {'dep': 'amod',
                                                  'dependent': 2,
                                                  'dependentGloss': 'quick',
                                                  'governor': 4,
                                                  'governorGloss': 'fox'},
                                                 {'dep': 'amod',
                                                  'dependent': 3,
                                                  'dependentGloss': 'brown',
                                                  'governor': 4,
                                                  'governorGloss': 'fox'},
                                                 {'dep': 'dep',
                                                  'dependent': 5,
                                                  'dependentGloss': 'jumps',
                                                  'governor': 4,
                                                  'governorGloss': 'fox'},
                                                 {'dep': 'case',
                                                  'dependent': 6,
                                                  'dependentGloss': 'over',
                                                  'governor': 9,
                                                  'governorGloss': 'dog'},
                                                 {'dep': 'det',
                                                  'dependent': 7,
                                                  'dependentGloss': 'the',
                                                  'governor': 9,
                                                  'governorGloss': 'dog'},
                                                 {'dep': 'amod',
                                                  'dependent': 8,
                                                  'dependentGloss': 'lazy',
                                                  'governor': 9,
                                                  'governorGloss': 'dog'},
                                                 {'dep': 'nmod',
                                                  'dependent': 9,
                                                  'dependentGloss': 'dog',
                                                  'governor': 5,
                                                  'governorGloss': 'jumps'}],
                           'enhancedDependencies': [{'dep': 'ROOT',
                                                     'dependent': 4,
                                                     'dependentGloss': 'fox',
                                                     'governor': 0,
                                                     'governorGloss': 'ROOT'},
                                                    {'dep': 'det',
                                                     'dependent': 1,
                                                     'dependentGloss': 'The',
                                                     'governor': 4,
                                                     'governorGloss': 'fox'},
                                                    {'dep': 'amod',
                                                     'dependent': 2,
                                                     'dependentGloss': 'quick',
                                                     'governor': 4,
                                                     'governorGloss': 'fox'},
                                                    {'dep': 'amod',
                                                     'dependent': 3,
                                                     'dependentGloss': 'brown',
                                                     'governor': 4,
                                                     'governorGloss': 'fox'},
                                                    {'dep': 'dep',
                                                     'dependent': 5,
                                                     'dependentGloss': 'jumps',
                                                     'governor': 4,
                                                     'governorGloss': 'fox'},
                                                    {'dep': 'case',
                                                     'dependent': 6,
                                                     'dependentGloss': 'over',
                                                     'governor': 9,
                                                     'governorGloss': 'dog'},
                                                    {'dep': 'det',
                                                     'dependent': 7,
                                                     'dependentGloss': 'the',
                                                     'governor': 9,
                                                     'governorGloss': 'dog'},
                                                    {'dep': 'amod',
                                                     'dependent': 8,
                                                     'dependentGloss': 'lazy',
                                                     'governor': 9,
                                                     'governorGloss': 'dog'},
                                                    {'dep': 'nmod:over',
                                                     'dependent': 9,
                                                     'dependentGloss': 'dog',
                                                     'governor': 5,
                                                     'governorGloss': 'jumps'}],
                           'enhancedPlusPlusDependencies': [{'dep': 'ROOT',
                                                             'dependent': 4,
                                                             'dependentGloss': 'fox',
                                                             'governor': 0,
                                                             'governorGloss': 'ROOT'},
                                                            {'dep': 'det',
                                                             'dependent': 1,
                                                             'dependentGloss': 'The',
                                                             'governor': 4,
                                                             'governorGloss': 'fox'},
                                                            {'dep': 'amod',
                                                             'dependent': 2,
                                                             'dependentGloss': 'quick',
                                                             'governor': 4,
                                                             'governorGloss': 'fox'},
                                                            {'dep': 'amod',
                                                             'dependent': 3,
                                                             'dependentGloss': 'brown',
                                                             'governor': 4,
                                                             'governorGloss': 'fox'},
                                                            {'dep': 'dep',
                                                             'dependent': 5,
                                                             'dependentGloss': 'jumps',
                                                             'governor': 4,
                                                             'governorGloss': 'fox'},
                                                            {'dep': 'case',
                                                             'dependent': 6,
                                                             'dependentGloss': 'over',
                                                             'governor': 9,
                                                             'governorGloss': 'dog'},
                                                            {'dep': 'det',
                                                             'dependent': 7,
                                                             'dependentGloss': 'the',
                                                             'governor': 9,
                                                             'governorGloss': 'dog'},
                                                            {'dep': 'amod',
                                                             'dependent': 8,
                                                             'dependentGloss': 'lazy',
                                                             'governor': 9,
                                                             'governorGloss': 'dog'},
                                                            {'dep': 'nmod:over',
                                                             'dependent': 9,
                                                             'dependentGloss': 'dog',
                                                             'governor': 5,
                                                             'governorGloss': 'jumps'}],
                           'index': 0,
                           'parse': '(ROOT\n  (NP\n    (NP (DT The) (JJ quick) (JJ brown) (NN fox))\n    (NP\n      (NP (NNS jumps))\n      (PP (IN over)\n        (NP (DT the) (JJ lazy) (NN dog))))))',
                           'tokens': [{'after': ' ',
                                       'before': '',
                                       'characterOffsetBegin': 0,
                                       'characterOffsetEnd': 3,
                                       'index': 1,
                                       'lemma': 'the',
                                       'originalText': 'The',
                                       'pos': 'DT',
                                       'word': 'The'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 4,
                                       'characterOffsetEnd': 9,
                                       'index': 2,
                                       'lemma': 'quick',
                                       'originalText': 'quick',
                                       'pos': 'JJ',
                                       'word': 'quick'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 10,
                                       'characterOffsetEnd': 15,
                                       'index': 3,
                                       'lemma': 'brown',
                                       'originalText': 'brown',
                                       'pos': 'JJ',
                                       'word': 'brown'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 16,
                                       'characterOffsetEnd': 19,
                                       'index': 4,
                                       'lemma': 'fox',
                                       'originalText': 'fox',
                                       'pos': 'NN',
                                       'word': 'fox'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 20,
                                       'characterOffsetEnd': 25,
                                       'index': 5,
                                       'lemma': 'jump',
                                       'originalText': 'jumps',
                                       'pos': 'VBZ',
                                       'word': 'jumps'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 26,
                                       'characterOffsetEnd': 30,
                                       'index': 6,
                                       'lemma': 'over',
                                       'originalText': 'over',
                                       'pos': 'IN',
                                       'word': 'over'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 31,
                                       'characterOffsetEnd': 34,
                                       'index': 7,
                                       'lemma': 'the',
                                       'originalText': 'the',
                                       'pos': 'DT',
                                       'word': 'the'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 35,
                                       'characterOffsetEnd': 39,
                                       'index': 8,
                                       'lemma': 'lazy',
                                       'originalText': 'lazy',
                                       'pos': 'JJ',
                                       'word': 'lazy'},
                                      {'after': '',
                                       'before': ' ',
                                       'characterOffsetBegin': 40,
                                       'characterOffsetEnd': 43,
                                       'index': 9,
                                       'lemma': 'dog',
                                       'originalText': 'dog',
                                       'pos': 'NN',
                                       'word': 'dog'}]}]
        }

        corenlp_parser.api_call = MagicMock(return_value=api_return_value)

        input_string = "The quick brown fox jumps over the lazy dog".split()
        expected_output = Tree('ROOT', [Tree('NP', [Tree('NP', [Tree('DT', ['The']), Tree('JJ', ['quick']),
                          Tree('JJ', ['brown']), Tree('NN', ['fox'])]), Tree('NP', [Tree('NP', [Tree('NNS', ['jumps'])]),
                          Tree('PP', [Tree('IN', ['over']), Tree('NP', [Tree('DT', ['the']), Tree('JJ', ['lazy']), Tree('NN', ['dog'])])])])])])

        parsed_data = next(corenlp_parser.parse(input_string))

        corenlp_parser.api_call.assert_called_once_with("The quick brown fox jumps over the lazy dog", properties={
                                                        'ssplit.ssplit.eolonly': 'true'})
        self.assertEqual(expected_output, parsed_data)

    def test_dependency_parser(self):
        corenlp_parser = corenlp.CoreNLPDependencyParser()

        api_return_value = {
            'sentences': [{'basicDependencies': [{'dep': 'ROOT',
                                                  'dependent': 5,
                                                  'dependentGloss': 'jumps',
                                                  'governor': 0,
                                                  'governorGloss': 'ROOT'},
                                                 {'dep': 'det',
                                                  'dependent': 1,
                                                  'dependentGloss': 'The',
                                                  'governor': 4,
                                                  'governorGloss': 'fox'},
                                                 {'dep': 'amod',
                                                  'dependent': 2,
                                                  'dependentGloss': 'quick',
                                                  'governor': 4,
                                                  'governorGloss': 'fox'},
                                                 {'dep': 'amod',
                                                  'dependent': 3,
                                                  'dependentGloss': 'brown',
                                                  'governor': 4,
                                                  'governorGloss': 'fox'},
                                                 {'dep': 'nsubj',
                                                  'dependent': 4,
                                                  'dependentGloss': 'fox',
                                                  'governor': 5,
                                                  'governorGloss': 'jumps'},
                                                 {'dep': 'case',
                                                  'dependent': 6,
                                                  'dependentGloss': 'over',
                                                  'governor': 9,
                                                  'governorGloss': 'dog'},
                                                 {'dep': 'det',
                                                  'dependent': 7,
                                                  'dependentGloss': 'the',
                                                  'governor': 9,
                                                  'governorGloss': 'dog'},
                                                 {'dep': 'amod',
                                                  'dependent': 8,
                                                  'dependentGloss': 'lazy',
                                                  'governor': 9,
                                                  'governorGloss': 'dog'},
                                                 {'dep': 'nmod',
                                                  'dependent': 9,
                                                  'dependentGloss': 'dog',
                                                  'governor': 5,
                                                  'governorGloss': 'jumps'}],
                           'enhancedDependencies': [{'dep': 'ROOT',
                                                     'dependent': 5,
                                                     'dependentGloss': 'jumps',
                                                     'governor': 0,
                                                     'governorGloss': 'ROOT'},
                                                    {'dep': 'det',
                                                     'dependent': 1,
                                                     'dependentGloss': 'The',
                                                     'governor': 4,
                                                     'governorGloss': 'fox'},
                                                    {'dep': 'amod',
                                                     'dependent': 2,
                                                     'dependentGloss': 'quick',
                                                     'governor': 4,
                                                     'governorGloss': 'fox'},
                                                    {'dep': 'amod',
                                                     'dependent': 3,
                                                     'dependentGloss': 'brown',
                                                     'governor': 4,
                                                     'governorGloss': 'fox'},
                                                    {'dep': 'nsubj',
                                                     'dependent': 4,
                                                     'dependentGloss': 'fox',
                                                     'governor': 5,
                                                     'governorGloss': 'jumps'},
                                                    {'dep': 'case',
                                                     'dependent': 6,
                                                     'dependentGloss': 'over',
                                                     'governor': 9,
                                                     'governorGloss': 'dog'},
                                                    {'dep': 'det',
                                                     'dependent': 7,
                                                     'dependentGloss': 'the',
                                                     'governor': 9,
                                                     'governorGloss': 'dog'},
                                                    {'dep': 'amod',
                                                     'dependent': 8,
                                                     'dependentGloss': 'lazy',
                                                     'governor': 9,
                                                     'governorGloss': 'dog'},
                                                    {'dep': 'nmod:over',
                                                     'dependent': 9,
                                                     'dependentGloss': 'dog',
                                                     'governor': 5,
                                                     'governorGloss': 'jumps'}],
                           'enhancedPlusPlusDependencies': [{'dep': 'ROOT',
                                                             'dependent': 5,
                                                             'dependentGloss': 'jumps',
                                                             'governor': 0,
                                                             'governorGloss': 'ROOT'},
                                                            {'dep': 'det',
                                                             'dependent': 1,
                                                             'dependentGloss': 'The',
                                                             'governor': 4,
                                                             'governorGloss': 'fox'},
                                                            {'dep': 'amod',
                                                             'dependent': 2,
                                                             'dependentGloss': 'quick',
                                                             'governor': 4,
                                                             'governorGloss': 'fox'},
                                                            {'dep': 'amod',
                                                             'dependent': 3,
                                                             'dependentGloss': 'brown',
                                                             'governor': 4,
                                                             'governorGloss': 'fox'},
                                                            {'dep': 'nsubj',
                                                             'dependent': 4,
                                                             'dependentGloss': 'fox',
                                                             'governor': 5,
                                                             'governorGloss': 'jumps'},
                                                            {'dep': 'case',
                                                             'dependent': 6,
                                                             'dependentGloss': 'over',
                                                             'governor': 9,
                                                             'governorGloss': 'dog'},
                                                            {'dep': 'det',
                                                             'dependent': 7,
                                                             'dependentGloss': 'the',
                                                             'governor': 9,
                                                             'governorGloss': 'dog'},
                                                            {'dep': 'amod',
                                                             'dependent': 8,
                                                             'dependentGloss': 'lazy',
                                                             'governor': 9,
                                                             'governorGloss': 'dog'},
                                                            {'dep': 'nmod:over',
                                                             'dependent': 9,
                                                             'dependentGloss': 'dog',
                                                             'governor': 5,
                                                             'governorGloss': 'jumps'}],
                           'index': 0,
                           'tokens': [{'after': ' ',
                                       'before': '',
                                       'characterOffsetBegin': 0,
                                       'characterOffsetEnd': 3,
                                       'index': 1,
                                       'lemma': 'the',
                                       'originalText': 'The',
                                       'pos': 'DT',
                                       'word': 'The'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 4,
                                       'characterOffsetEnd': 9,
                                       'index': 2,
                                       'lemma': 'quick',
                                       'originalText': 'quick',
                                       'pos': 'JJ',
                                       'word': 'quick'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 10,
                                       'characterOffsetEnd': 15,
                                       'index': 3,
                                       'lemma': 'brown',
                                       'originalText': 'brown',
                                       'pos': 'JJ',
                                       'word': 'brown'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 16,
                                       'characterOffsetEnd': 19,
                                       'index': 4,
                                       'lemma': 'fox',
                                       'originalText': 'fox',
                                       'pos': 'NN',
                                       'word': 'fox'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 20,
                                       'characterOffsetEnd': 25,
                                       'index': 5,
                                       'lemma': 'jump',
                                       'originalText': 'jumps',
                                       'pos': 'VBZ',
                                       'word': 'jumps'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 26,
                                       'characterOffsetEnd': 30,
                                       'index': 6,
                                       'lemma': 'over',
                                       'originalText': 'over',
                                       'pos': 'IN',
                                       'word': 'over'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 31,
                                       'characterOffsetEnd': 34,
                                       'index': 7,
                                       'lemma': 'the',
                                       'originalText': 'the',
                                       'pos': 'DT',
                                       'word': 'the'},
                                      {'after': ' ',
                                       'before': ' ',
                                       'characterOffsetBegin': 35,
                                       'characterOffsetEnd': 39,
                                       'index': 8,
                                       'lemma': 'lazy',
                                       'originalText': 'lazy',
                                       'pos': 'JJ',
                                       'word': 'lazy'},
                                      {'after': '',
                                       'before': ' ',
                                       'characterOffsetBegin': 40,
                                       'characterOffsetEnd': 43,
                                       'index': 9,
                                       'lemma': 'dog',
                                       'originalText': 'dog',
                                       'pos': 'NN',
                                       'word': 'dog'}]}]
        }

        corenlp_parser.api_call = MagicMock(return_value=api_return_value)

        input_string = "The quick brown fox jumps over the lazy dog".split()
        expected_output = Tree('jumps', [Tree('fox', ['The', 'quick', 'brown']), Tree(
            'dog', ['over', 'the', 'lazy'])])

        parsed_data = next(corenlp_parser.parse(input_string))

        corenlp_parser.api_call.assert_called_once_with("The quick brown fox jumps over the lazy dog", properties={
                                                        'ssplit.ssplit.eolonly': 'true'})
        self.assertEqual(expected_output, parsed_data.tree())



# Natural Language Toolkit: Classifier Interface
#
# Copyright (C) 2001-2018 NLTK Project
# Author: Edward Loper <edloper@gmail.com>
#         Steven Bird <stevenbird1@gmail.com> (minor additions)
# URL: <http://nltk.org/>
# For license information, see LICENSE.TXT

"""
Interfaces for labeling tokens with category labels (or "class labels").

``ClassifierI`` is a standard interface for "single-category
classification", in which the set of categories is known, the number
of categories is finite, and each text belongs to exactly one
category.

``MultiClassifierI`` is a standard interface for "multi-category
classification", which is like single-category classification except
that each text belongs to zero or more categories.
"""
from nltk.internals import overridden

##//////////////////////////////////////////////////////
#{ Classification Interfaces
##//////////////////////////////////////////////////////

class ClassifierI(object):
    """
    A processing interface for labeling tokens with a single category
    label (or "class").  Labels are typically strs or
    ints, but can be any immutable type.  The set of labels
    that the classifier chooses from must be fixed and finite.

    Subclasses must define:
      - ``labels()``
      - either ``classify()`` or ``classify_many()`` (or both)

    Subclasses may define:
      - either ``prob_classify()`` or ``prob_classify_many()`` (or both)
    """
    def labels(self):
        """
        :return: the list of category labels used by this classifier.
        :rtype: list of (immutable)
        """
        raise NotImplementedError()

    def classify(self, featureset):
        """
        :return: the most appropriate label for the given featureset.
        :rtype: label
        """
        if overridden(self.classify_many):
            return self.classify_many([featureset])[0]
        else:
            raise NotImplementedError()

    def prob_classify(self, featureset):
        """
        :return: a probability distribution over labels for the given
            featureset.
        :rtype: ProbDistI
        """
        if overridden(self.prob_classify_many):
            return self.prob_classify_many([featureset])[0]
        else:
            raise NotImplementedError()

    def classify_many(self, featuresets):
        """
        Apply ``self.classify()`` to each element of ``featuresets``.  I.e.:

            return [self.classify(fs) for fs in featuresets]

        :rtype: list(label)
        """
        return [self.classify(fs) for fs in featuresets]

    def prob_classify_many(self, featuresets):
        """
        Apply ``self.prob_classify()`` to each element of ``featuresets``.  I.e.:

            return [self.prob_classify(fs) for fs in featuresets]

        :rtype: list(ProbDistI)
        """
        return [self.prob_classify(fs) for fs in featuresets]


class MultiClassifierI(object):
    """
    A processing interface for labeling tokens with zero or more
    category labels (or "labels").  Labels are typically strs
    or ints, but can be any immutable type.  The set of labels
    that the multi-classifier chooses from must be fixed and finite.

    Subclasses must define:
      - ``labels()``
      - either ``classify()`` or ``classify_many()`` (or both)

    Subclasses may define:
      - either ``prob_classify()`` or ``prob_classify_many()`` (or both)
    """
    def labels(self):
        """
        :return: the list of category labels used by this classifier.
        :rtype: list of (immutable)
        """
        raise NotImplementedError()

    def classify(self, featureset):
        """
        :return: the most appropriate set of labels for the given featureset.
        :rtype: set(label)
        """
        if overridden(self.classify_many):
            return self.classify_many([featureset])[0]
        else:
            raise NotImplementedError()

    def prob_classify(self, featureset):
        """
        :return: a probability distribution over sets of labels for the
            given featureset.
        :rtype: ProbDistI
        """
        if overridden(self.prob_classify_many):
            return self.prob_classify_many([featureset])[0]
        else:
            raise NotImplementedError()

    def classify_many(self, featuresets):
        """
        Apply ``self.classify()`` to each element of ``featuresets``.  I.e.:

            return [self.classify(fs) for fs in featuresets]

        :rtype: list(set(label))
        """
        return [self.classify(fs) for fs in featuresets]

    def prob_classify_many(self, featuresets):
        """
        Apply ``self.prob_classify()`` to each element of ``featuresets``.  I.e.:

            return [self.prob_classify(fs) for fs in featuresets]

        :rtype: list(ProbDistI)
        """
        return [self.prob_classify(fs) for fs in featuresets]


# # [XX] IN PROGRESS:
# class SequenceClassifierI(object):
#     """
#     A processing interface for labeling sequences of tokens with a
#     single category label (or "class").  Labels are typically
#     strs or ints, but can be any immutable type.  The set
#     of labels that the classifier chooses from must be fixed and
#     finite.
#     """
#     def labels(self):
#         """
#         :return: the list of category labels used by this classifier.
#         :rtype: list of (immutable)
#         """
#         raise NotImplementedError()

#     def prob_classify(self, featureset):
#         """
#         Return a probability distribution over labels for the given
#         featureset.

#         If ``featureset`` is a list of featuresets, then return a
#         corresponding list containing the probability distribution
#         over labels for each of the given featuresets, where the
#         *i*\ th element of this list is the most appropriate label for
#         the *i*\ th element of ``featuresets``.
#         """
#         raise NotImplementedError()

#     def classify(self, featureset):
#         """
#         Return the most appropriate label for the given featureset.

#         If ``featureset`` is a list of featuresets, then return a
#         corresponding list containing the most appropriate label for
#         each of the given featuresets, where the *i*\ th element of
#         this list is the most appropriate label for the *i*\ th element
#         of ``featuresets``.
#         """
#         raise NotImplementedError()




# Natural Language Toolkit: Penn Treebank Reader
#
# Copyright (C) 2001-2018 NLTK Project
# Author: Steven Bird <stevenbird1@gmail.com>
#         Edward Loper <edloper@gmail.com>
# URL: <http://nltk.org/>
# For license information, see LICENSE.TXT
"""
Corpus reader for corpora that consist of parenthesis-delineated parse trees.
"""

import sys

from nltk.tree import Tree
from nltk.tag import map_tag

from nltk.corpus.reader.util import *
from nltk.corpus.reader.api import *

# we use [^\s()]+ instead of \S+? to avoid matching ()
SORTTAGWRD = re.compile(r'\((\d+) ([^\s()]+) ([^\s()]+)\)') 
TAGWORD = re.compile(r'\(([^\s()]+) ([^\s()]+)\)')
WORD = re.compile(r'\([^\s()]+ ([^\s()]+)\)')
EMPTY_BRACKETS = re.compile(r'\s*\(\s*\(')

class BracketParseCorpusReader(SyntaxCorpusReader):
    """
    Reader for corpora that consist of parenthesis-delineated parse trees,
    like those found in the "combined" section of the Penn Treebank,
    e.g. "(S (NP (DT the) (JJ little) (NN dog)) (VP (VBD barked)))".

    """
    def __init__(self, root, fileids, comment_char=None,
                 detect_blocks='unindented_paren', encoding='utf8',
                 tagset=None):
        """
        :param root: The root directory for this corpus.
        :param fileids: A list or regexp specifying the fileids in this corpus.
        :param comment_char: The character which can appear at the start of
            a line to indicate that the rest of the line is a comment.
        :param detect_blocks: The method that is used to find blocks
          in the corpus; can be 'unindented_paren' (every unindented
          parenthesis starts a new parse) or 'sexpr' (brackets are
          matched).
        :param tagset: The name of the tagset used by this corpus, to be used
              for normalizing or converting the POS tags returned by the
              tagged_...() methods.
        """
        CorpusReader.__init__(self, root, fileids, encoding)
        self._comment_char = comment_char
        self._detect_blocks = detect_blocks
        self._tagset = tagset

    def _read_block(self, stream):
        if self._detect_blocks == 'sexpr':
            return read_sexpr_block(stream, comment_char=self._comment_char)
        elif self._detect_blocks == 'blankline':
            return read_blankline_block(stream)
        elif self._detect_blocks == 'unindented_paren':
            # Tokens start with unindented left parens.
            toks = read_regexp_block(stream, start_re=r'^\(')
            # Strip any comments out of the tokens.
            if self._comment_char:
                toks = [re.sub('(?m)^%s.*'%re.escape(self._comment_char),
                               '', tok)
                        for tok in toks]
            return toks
        else:
            assert 0, 'bad block type'

    def _normalize(self, t):
        # If there's an empty set of brackets surrounding the actual
        # parse, then strip them off.
        if EMPTY_BRACKETS.match(t):
            t = t.strip()[1:-1]
        # Replace leaves of the form (!), (,), with (! !), (, ,)
        t = re.sub(r"\((.)\)", r"(\1 \1)", t)
        # Replace leaves of the form (tag word root) with (tag word)
        t = re.sub(r"\(([^\s()]+) ([^\s()]+) [^\s()]+\)", r"(\1 \2)", t)
        return t

    def _parse(self, t):
        try:
            return Tree.fromstring(self._normalize(t))

        except ValueError as e:
            sys.stderr.write("Bad tree detected; trying to recover...\n")
            # Try to recover, if we can:
            if e.args == ('mismatched parens',):
                for n in range(1, 5):
                    try:
                        v = Tree(self._normalize(t+')'*n))
                        sys.stderr.write("  Recovered by adding %d close "
                                         "paren(s)\n" % n)
                        return v
                    except ValueError: pass
            # Try something else:
            sys.stderr.write("  Recovered by returning a flat parse.\n")
            #sys.stderr.write(' '.join(t.split())+'\n')
            return Tree('S', self._tag(t))

    def _tag(self, t, tagset=None):
        tagged_sent = [(w,p) for (p,w) in TAGWORD.findall(self._normalize(t))]
        if tagset and tagset != self._tagset:
            tagged_sent = [(w, map_tag(self._tagset, tagset, p)) for (w,p) in tagged_sent]
        return tagged_sent

    def _word(self, t):
        return WORD.findall(self._normalize(t))

class CategorizedBracketParseCorpusReader(CategorizedCorpusReader,
                                          BracketParseCorpusReader):
    """
    A reader for parsed corpora whose documents are
    divided into categories based on their file identifiers.
    @author: Nathan Schneider <nschneid@cs.cmu.edu>
    """
    def __init__(self, *args, **kwargs):
        """
        Initialize the corpus reader.  Categorization arguments
        (C{cat_pattern}, C{cat_map}, and C{cat_file}) are passed to
        the L{CategorizedCorpusReader constructor
        <CategorizedCorpusReader.__init__>}.  The remaining arguments
        are passed to the L{BracketParseCorpusReader constructor
        <BracketParseCorpusReader.__init__>}.
        """
        CategorizedCorpusReader.__init__(self, kwargs)
        BracketParseCorpusReader.__init__(self, *args, **kwargs)

    def _resolve(self, fileids, categories):
        if fileids is not None and categories is not None:
            raise ValueError('Specify fileids or categories, not both')
        if categories is not None:
            return self.fileids(categories)
        else:
            return fileids
    def raw(self, fileids=None, categories=None):
        return BracketParseCorpusReader.raw(
            self, self._resolve(fileids, categories))
    def words(self, fileids=None, categories=None):
        return BracketParseCorpusReader.words(
            self, self._resolve(fileids, categories))
    def sents(self, fileids=None, categories=None):
        return BracketParseCorpusReader.sents(
            self, self._resolve(fileids, categories))
    def paras(self, fileids=None, categories=None):
        return BracketParseCorpusReader.paras(
            self, self._resolve(fileids, categories))
    def tagged_words(self, fileids=None, categories=None, tagset=None):
        return BracketParseCorpusReader.tagged_words(
            self, self._resolve(fileids, categories), tagset)
    def tagged_sents(self, fileids=None, categories=None, tagset=None):
        return BracketParseCorpusReader.tagged_sents(
            self, self._resolve(fileids, categories), tagset)
    def tagged_paras(self, fileids=None, categories=None, tagset=None):
        return BracketParseCorpusReader.tagged_paras(
            self, self._resolve(fileids, categories), tagset)
    def parsed_words(self, fileids=None, categories=None):
        return BracketParseCorpusReader.parsed_words(
            self, self._resolve(fileids, categories))
    def parsed_sents(self, fileids=None, categories=None):
        return BracketParseCorpusReader.parsed_sents(
            self, self._resolve(fileids, categories))
    def parsed_paras(self, fileids=None, categories=None):
        return BracketParseCorpusReader.parsed_paras(
            self, self._resolve(fileids, categories))

class AlpinoCorpusReader(BracketParseCorpusReader):
    """
    Reader for the Alpino Dutch Treebank.
    This corpus has a lexical breakdown structure embedded, as read by _parse
    Unfortunately this puts punctuation and some other words out of the sentence
    order in the xml element tree. This is no good for tag_ and word_
    _tag and _word will be overridden to use a non-default new parameter 'ordered'
    to the overridden _normalize function. The _parse function can then remain 
    untouched.
    """
    def __init__(self, root, encoding='ISO-8859-1', tagset=None):
        BracketParseCorpusReader.__init__(self, root, 'alpino\.xml',
                                 detect_blocks='blankline',
                                 encoding=encoding,
                                 tagset=tagset)

    def _normalize(self, t, ordered = False):
        """Normalize the xml sentence element in t.
        The sentence elements <alpino_ds>, although embedded in a few overall 
        xml elements, are seperated by blank lines. That's how the reader can 
        deliver them one at a time.
        Each sentence has a few category subnodes that are of no use to us.
        The remaining word nodes may or may not appear in the proper order.
        Each word node has attributes, among which:
        - begin : the position of the word in the sentence
        - pos   : Part of Speech: the Tag
        - word  : the actual word
        The return value is a string with all xml elementes replaced by 
        clauses: either a cat clause with nested clauses, or a word clause.
        The order of the bracket clauses closely follows the xml.
        If ordered == True, the word clauses include an order sequence number.
        If ordered == False, the word clauses only have pos and word parts.
        """
        if t[:10] != "<alpino_ds":
            return ""
        # convert XML to sexpr notation
        t = re.sub(r'  <node .*? cat="(\w+)".*>', r"(\1", t)
        if ordered:
            t = re.sub(r'  <node. *?begin="(\d+)".*? pos="(\w+)".*? word="([^"]+)".*?/>', r"(\1 \2 \3)", t)
        else: 
            t = re.sub(r'  <node .*?pos="(\w+)".*? word="([^"]+)".*?/>', r"(\1 \2)", t)
        t = re.sub(r"  </node>", r")", t)
        t = re.sub(r"<sentence>.*</sentence>", r"", t)
        t = re.sub(r"</?alpino_ds.*>", r"", t)
        return t

    def _tag(self, t, tagset=None):
        tagged_sent = [(int(o), w, p) for (o,p,w) in SORTTAGWRD.findall(self._normalize(t, ordered = True))]
        tagged_sent.sort()
        if tagset and tagset != self._tagset:
            tagged_sent = [(w, map_tag(self._tagset, tagset, p)) for (o,w,p) in tagged_sent]
        else:
            tagged_sent = [(w,p) for (o,w,p) in tagged_sent]
        return tagged_sent

    def _word(self, t):
        """Return a correctly ordered list if words"""
        tagged_sent = self._tag(t)
        return [w for (w,p) in tagged_sent]      




# -*- coding: utf-8 -*-
"""
Unit tests for Senna
"""

from __future__ import unicode_literals
from os import environ, path, sep

import logging
import unittest

from nltk.classify import Senna
from nltk.tag import SennaTagger, SennaChunkTagger, SennaNERTagger

# Set Senna executable path for tests if it is not specified as an environment variable
if 'SENNA' in environ:
    SENNA_EXECUTABLE_PATH = path.normpath(environ['SENNA']) + sep
else:
    SENNA_EXECUTABLE_PATH = '/usr/share/senna-v3.0'

senna_is_installed = path.exists(SENNA_EXECUTABLE_PATH)

@unittest.skipUnless(senna_is_installed, "Requires Senna executable")
class TestSennaPipeline(unittest.TestCase):
    """Unittest for nltk.classify.senna"""

    def test_senna_pipeline(self):
        """Senna pipeline interface"""

        pipeline = Senna(SENNA_EXECUTABLE_PATH, ['pos', 'chk', 'ner'])
        sent = 'Dusseldorf is an international business center'.split()
        result = [(token['word'], token['chk'], token['ner'], token['pos']) for token in pipeline.tag(sent)]
        expected = [('Dusseldorf', 'B-NP', 'B-LOC', 'NNP'), ('is', 'B-VP',
            'O', 'VBZ'), ('an', 'B-NP', 'O', 'DT'), ('international', 'I-NP',
            'O', 'JJ'), ('business', 'I-NP', 'O', 'NN'), ('center', 'I-NP',
            'O', 'NN')]
        self.assertEqual(result, expected)

@unittest.skipUnless(senna_is_installed, "Requires Senna executable")
class TestSennaTagger(unittest.TestCase):
    """Unittest for nltk.tag.senna"""

    def test_senna_tagger(self):
        tagger = SennaTagger(SENNA_EXECUTABLE_PATH)
        result = tagger.tag('What is the airspeed of an unladen swallow ?'.split())
        expected = [('What', 'WP'), ('is', 'VBZ'), ('the', 'DT'), ('airspeed',
            'NN'),('of', 'IN'), ('an', 'DT'), ('unladen', 'NN'), ('swallow',
            'NN'), ('?', '.')]
        self.assertEqual(result, expected)

    def test_senna_chunk_tagger(self):
        chktagger = SennaChunkTagger(SENNA_EXECUTABLE_PATH)
        result_1 = chktagger.tag('What is the airspeed of an unladen swallow ?'.split())
        expected_1 = [('What', 'B-NP'), ('is', 'B-VP'), ('the', 'B-NP'), ('airspeed',
            'I-NP'), ('of', 'B-PP'), ('an', 'B-NP'), ('unladen', 'I-NP'), ('swallow',
            'I-NP'), ('?', 'O')]

        result_2 = list(chktagger.bio_to_chunks(result_1, chunk_type='NP'))
        expected_2 = [('What', '0'), ('the airspeed', '2-3'), ('an unladen swallow',
            '5-6-7')]
        self.assertEqual(result_1, expected_1)
        self.assertEqual(result_2, expected_2)

    def test_senna_ner_tagger(self):
        nertagger = SennaNERTagger(SENNA_EXECUTABLE_PATH)
        result_1 = nertagger.tag('Shakespeare theatre was in London .'.split())
        expected_1 = [('Shakespeare', 'B-PER'), ('theatre', 'O'), ('was', 'O'),
                ('in', 'O'), ('London', 'B-LOC'), ('.', 'O')]

        result_2 = nertagger.tag('UN headquarters are in NY , USA .'.split())
        expected_2 = [('UN', 'B-ORG'), ('headquarters', 'O'), ('are', 'O'),
        ('in', 'O'), ('NY', 'B-LOC'), (',', 'O'), ('USA', 'B-LOC'), ('.', 'O')]
        self.assertEqual(result_1, expected_1)
        self.assertEqual(result_2, expected_2)



# Natural Language Toolkit: Tokenizers
#
# Copyright (C) 2001-2018 NLTK Project
# Author: Yoav Goldberg <yoavg@cs.bgu.ac.il>
#         Steven Bird <stevenbird1@gmail.com> (minor edits)
# URL: <http://nltk.sourceforge.net>
# For license information, see LICENSE.TXT

"""
S-Expression Tokenizer

``SExprTokenizer`` is used to find parenthesized expressions in a
string.  In particular, it divides a string into a sequence of
substrings that are either parenthesized expressions (including any
nested parenthesized expressions), or other whitespace-separated
tokens.

    >>> from nltk.tokenize import SExprTokenizer
    >>> SExprTokenizer().tokenize('(a b (c d)) e f (g)')
    ['(a b (c d))', 'e', 'f', '(g)']

By default, `SExprTokenizer` will raise a ``ValueError`` exception if
used to tokenize an expression with non-matching parentheses:

    >>> SExprTokenizer().tokenize('c) d) e (f (g')
    Traceback (most recent call last):
      ...
    ValueError: Un-matched close paren at char 1

The ``strict`` argument can be set to False to allow for
non-matching parentheses.  Any unmatched close parentheses will be
listed as their own s-expression; and the last partial sexpr with
unmatched open parentheses will be listed as its own sexpr:

    >>> SExprTokenizer(strict=False).tokenize('c) d) e (f (g')
    ['c', ')', 'd', ')', 'e', '(f (g']

The characters used for open and close parentheses may be customized
using the ``parens`` argument to the `SExprTokenizer` constructor:

    >>> SExprTokenizer(parens='{}').tokenize('{a b {c d}} e f {g}')
    ['{a b {c d}}', 'e', 'f', '{g}']

The s-expression tokenizer is also available as a function:

    >>> from nltk.tokenize import sexpr_tokenize
    >>> sexpr_tokenize('(a b (c d)) e f (g)')
    ['(a b (c d))', 'e', 'f', '(g)']

"""

import re

from nltk.tokenize.api import TokenizerI


class SExprTokenizer(TokenizerI):
    """
    A tokenizer that divides strings into s-expressions.
    An s-expresion can be either:

      - a parenthesized expression, including any nested parenthesized
        expressions, or
      - a sequence of non-whitespace non-parenthesis characters.

    For example, the string ``(a (b c)) d e (f)`` consists of four
    s-expressions: ``(a (b c))``, ``d``, ``e``, and ``(f)``.

    By default, the characters ``(`` and ``)`` are treated as open and
    close parentheses, but alternative strings may be specified.

    :param parens: A two-element sequence specifying the open and close parentheses
        that should be used to find sexprs.  This will typically be either a
        two-character string, or a list of two strings.
    :type parens: str or list
    :param strict: If true, then raise an exception when tokenizing an ill-formed sexpr.
    """

    def __init__(self, parens='()', strict=True):
        if len(parens) != 2:
            raise ValueError('parens must contain exactly two strings')
        self._strict = strict
        self._open_paren = parens[0]
        self._close_paren = parens[1]
        self._paren_regexp = re.compile('%s|%s' % (re.escape(parens[0]),
                                                   re.escape(parens[1])))

    def tokenize(self, text):
        """
        Return a list of s-expressions extracted from *text*.
        For example:

            >>> SExprTokenizer().tokenize('(a b (c d)) e f (g)')
            ['(a b (c d))', 'e', 'f', '(g)']

        All parentheses are assumed to mark s-expressions.
        (No special processing is done to exclude parentheses that occur
        inside strings, or following backslash characters.)

        If the given expression contains non-matching parentheses,
        then the behavior of the tokenizer depends on the ``strict``
        parameter to the constructor.  If ``strict`` is ``True``, then
        raise a ``ValueError``.  If ``strict`` is ``False``, then any
        unmatched close parentheses will be listed as their own
        s-expression; and the last partial s-expression with unmatched open
        parentheses will be listed as its own s-expression:

            >>> SExprTokenizer(strict=False).tokenize('c) d) e (f (g')
            ['c', ')', 'd', ')', 'e', '(f (g']

        :param text: the string to be tokenized
        :type text: str or iter(str)
        :rtype: iter(str)
        """
        result = []
        pos = 0
        depth = 0
        for m in self._paren_regexp.finditer(text):
            paren = m.group()
            if depth == 0:
                result += text[pos:m.start()].split()
                pos = m.start()
            if paren == self._open_paren:
                depth += 1
            if paren == self._close_paren:
                if self._strict and depth == 0:
                    raise ValueError('Un-matched close paren at char %d'
                                     % m.start())
                depth = max(0, depth - 1)
                if depth == 0:
                    result.append(text[pos:m.end()])
                    pos = m.end()
        if self._strict and depth > 0:
            raise ValueError('Un-matched open paren at char %d' % pos)
        if pos < len(text):
            result.append(text[pos:])
        return result


sexpr_tokenize = SExprTokenizer().tokenize



# -*- coding: utf-8 -*-
# Natural Language Toolkit: Generating from a CFG
#
# Copyright (C) 2001-2018 NLTK Project
# Author: Steven Bird <stevenbird1@gmail.com>
#         Peter Ljunglöf <peter.ljunglof@heatherleaf.se>
# URL: <http://nltk.org/>
# For license information, see LICENSE.TXT
#
from __future__ import print_function

import itertools
import sys
from nltk.grammar import Nonterminal


def generate(grammar, start=None, depth=None, n=None):
    """
    Generates an iterator of all sentences from a CFG.

    :param grammar: The Grammar used to generate sentences.
    :param start: The Nonterminal from which to start generate sentences.
    :param depth: The maximal depth of the generated tree.
    :param n: The maximum number of sentences to return.
    :return: An iterator of lists of terminal tokens.
    """
    if not start:
        start = grammar.start()
    if depth is None:
        depth = sys.maxsize

    iter = _generate_all(grammar, [start], depth)

    if n:
        iter = itertools.islice(iter, n)

    return iter


def _generate_all(grammar, items, depth):
    if items:
        try:
            for frag1 in _generate_one(grammar, items[0], depth):
                for frag2 in _generate_all(grammar, items[1:], depth):
                    yield frag1 + frag2
        except RuntimeError as _error:
            if _error.message == "maximum recursion depth exceeded":
                # Helpful error message while still showing the recursion stack.
                raise RuntimeError("The grammar has rule(s) that yield infinite recursion!!")
            else:
                raise
    else:
        yield []


def _generate_one(grammar, item, depth):
    if depth > 0:
        if isinstance(item, Nonterminal):
            for prod in grammar.productions(lhs=item):
                for frag in _generate_all(grammar, prod.rhs(), depth-1):
                    yield frag
        else:
            yield [item]

demo_grammar = """
  S -> NP VP
  NP -> Det N
  PP -> P NP
  VP -> 'slept' | 'saw' NP | 'walked' PP
  Det -> 'the' | 'a'
  N -> 'man' | 'park' | 'dog'
  P -> 'in' | 'with'
"""


def demo(N=23):
    from nltk.grammar import CFG

    print('Generating the first %d sentences for demo grammar:' % (N,))
    print(demo_grammar)
    grammar = CFG.fromstring(demo_grammar)
    for n, sent in enumerate(generate(grammar, n=N), 1):
        print('%3d. %s' % (n, ' '.join(sent)))


if __name__ == '__main__':
    demo()



# Natural Language Toolkit: Interface to the Prover9 Theorem Prover
#
# Copyright (C) 2001-2018 NLTK Project
# Author: Dan Garrette <dhgarrette@gmail.com>
#         Ewan Klein <ewan@inf.ed.ac.uk>
#
# URL: <http://nltk.org/>
# For license information, see LICENSE.TXT
"""
A theorem prover that makes use of the external 'Prover9' package.
"""
from __future__ import print_function

import os
import subprocess

import nltk
from nltk.sem.logic import Expression, ExistsExpression, AllExpression, \
    NegatedExpression, AndExpression, IffExpression, OrExpression, \
    EqualityExpression, ImpExpression
from nltk.inference.api import BaseProverCommand, Prover

#
# Following is not yet used. Return code for 2 actually realized as 512.
#
p9_return_codes = {
    0: True,
    1:  "(FATAL)",      #A fatal error occurred (user's syntax error).
    2: False,           # (SOS_EMPTY) Prover9 ran out of things to do
                        #   (sos list exhausted).
    3: "(MAX_MEGS)",    # The max_megs (memory limit) parameter was exceeded.
    4: "(MAX_SECONDS)", # The max_seconds parameter was exceeded.
    5: "(MAX_GIVEN)",   # The max_given parameter was exceeded.
    6: "(MAX_KEPT)",    # The max_kept parameter was exceeded.
    7: "(ACTION)",      # A Prover9 action terminated the search.
    101: "(SIGSEGV)",   # Prover9 crashed, most probably due to a bug.
 }


class Prover9CommandParent(object):
    """
    A common base class used by both ``Prover9Command`` and ``MaceCommand``,
    which is responsible for maintaining a goal and a set of assumptions,
    and generating prover9-style input files from them.
    """
    def print_assumptions(self, output_format='nltk'):
        """
        Print the list of the current assumptions.
        """
        if output_format.lower() == 'nltk':
            for a in self.assumptions():
                print(a)
        elif output_format.lower() == 'prover9':
            for a in convert_to_prover9(self.assumptions()):
                print(a)
        else:
            raise NameError("Unrecognized value for 'output_format': %s" %
                            output_format)

class Prover9Command(Prover9CommandParent, BaseProverCommand):
    """
    A ``ProverCommand`` specific to the ``Prover9`` prover.  It contains
    the a print_assumptions() method that is used to print the list
    of assumptions in multiple formats.
    """
    def __init__(self, goal=None, assumptions=None, timeout=60, prover=None):
        """
        :param goal: Input expression to prove
        :type goal: sem.Expression
        :param assumptions: Input expressions to use as assumptions in
            the proof.
        :type assumptions: list(sem.Expression)
        :param timeout: number of seconds before timeout; set to 0 for
            no timeout.
        :type timeout: int
        :param prover: a prover.  If not set, one will be created.
        :type prover: Prover9
        """
        if not assumptions:
            assumptions = []

        if prover is not None:
            assert isinstance(prover, Prover9)
        else:
            prover = Prover9(timeout)

        BaseProverCommand.__init__(self, prover, goal, assumptions)

    def decorate_proof(self, proof_string, simplify=True):
        """
        :see BaseProverCommand.decorate_proof()
        """
        if simplify:
            return self._prover._call_prooftrans(proof_string, ['striplabels'])[0].rstrip()
        else:
            return proof_string.rstrip()


class Prover9Parent(object):
    """
    A common class extended by both ``Prover9`` and ``Mace <mace.Mace>``.
    It contains the functionality required to convert NLTK-style
    expressions into Prover9-style expressions.
    """

    _binary_location = None

    def config_prover9(self, binary_location, verbose=False):
        if binary_location is None:
            self._binary_location = None
            self._prover9_bin = None
        else:
            name = 'prover9'
            self._prover9_bin = nltk.internals.find_binary(
                                  name,
                                  path_to_bin=binary_location,
                                  env_vars=['PROVER9'],
                                  url='http://www.cs.unm.edu/~mccune/prover9/',
                                  binary_names=[name, name + '.exe'],
                                  verbose=verbose)
            self._binary_location = self._prover9_bin.rsplit(os.path.sep, 1)

    def prover9_input(self, goal, assumptions):
        """
        :return: The input string that should be provided to the
        prover9 binary.  This string is formed based on the goal,
        assumptions, and timeout value of this object.
        """
        s = ''

        if assumptions:
            s += 'formulas(assumptions).\n'
            for p9_assumption in convert_to_prover9(assumptions):
                s += '    %s.\n' % p9_assumption
            s += 'end_of_list.\n\n'

        if goal:
            s += 'formulas(goals).\n'
            s += '    %s.\n' % convert_to_prover9(goal)
            s += 'end_of_list.\n\n'

        return s

    def binary_locations(self):
        """
        A list of directories that should be searched for the prover9
        executables.  This list is used by ``config_prover9`` when searching
        for the prover9 executables.
        """
        return ['/usr/local/bin/prover9',
                '/usr/local/bin/prover9/bin',
                '/usr/local/bin',
                '/usr/bin',
                '/usr/local/prover9',
                '/usr/local/share/prover9']

    def _find_binary(self, name, verbose=False):
        binary_locations = self.binary_locations()
        if self._binary_location is not None:
            binary_locations += [self._binary_location]
        return nltk.internals.find_binary(name,
            searchpath=binary_locations,
            env_vars=['PROVER9'],
            url='http://www.cs.unm.edu/~mccune/prover9/',
            binary_names=[name, name + '.exe'],
            verbose=verbose)

    def _call(self, input_str, binary, args=[], verbose=False):
        """
        Call the binary with the given input.

        :param input_str: A string whose contents are used as stdin.
        :param binary: The location of the binary to call
        :param args: A list of command-line arguments.
        :return: A tuple (stdout, returncode)
        :see: ``config_prover9``
        """
        if verbose:
            print('Calling:', binary)
            print('Args:', args)
            print('Input:\n', input_str, '\n')

        # Call prover9 via a subprocess
        cmd = [binary] + args
        try:
            input_str = input_str.encode("utf8")
        except AttributeError:
            pass
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT,
                             stdin=subprocess.PIPE)
        (stdout, stderr) = p.communicate(input=input_str)

        if verbose:
            print('Return code:', p.returncode)
            if stdout: print('stdout:\n', stdout, '\n')
            if stderr: print('stderr:\n', stderr, '\n')

        return (stdout.decode("utf-8"), p.returncode)


def convert_to_prover9(input):
    """
    Convert a ``logic.Expression`` to Prover9 format.
    """
    if isinstance(input, list):
        result = []
        for s in input:
            try:
                result.append(_convert_to_prover9(s.simplify()))
            except:
                print('input %s cannot be converted to Prover9 input syntax' % input)
                raise
        return result
    else:
        try:
            return _convert_to_prover9(input.simplify())
        except:
            print('input %s cannot be converted to Prover9 input syntax' % input)
            raise

def _convert_to_prover9(expression):
    """
    Convert ``logic.Expression`` to Prover9 formatted string.
    """
    if isinstance(expression, ExistsExpression):
        return 'exists ' + str(expression.variable) + ' ' + _convert_to_prover9(expression.term)
    elif isinstance(expression, AllExpression):
        return 'all ' + str(expression.variable) + ' ' + _convert_to_prover9(expression.term)
    elif isinstance(expression, NegatedExpression):
        return '-(' + _convert_to_prover9(expression.term) + ')'
    elif isinstance(expression, AndExpression):
        return '(' + _convert_to_prover9(expression.first) + ' & ' + \
                     _convert_to_prover9(expression.second) + ')'
    elif isinstance(expression, OrExpression):
        return '(' + _convert_to_prover9(expression.first) + ' | ' + \
                     _convert_to_prover9(expression.second) + ')'
    elif isinstance(expression, ImpExpression):
        return '(' + _convert_to_prover9(expression.first) + ' -> ' + \
                     _convert_to_prover9(expression.second) + ')'
    elif isinstance(expression, IffExpression):
        return '(' + _convert_to_prover9(expression.first) + ' <-> ' + \
                     _convert_to_prover9(expression.second) + ')'
    elif isinstance(expression, EqualityExpression):
        return '(' + _convert_to_prover9(expression.first) + ' = ' + \
                     _convert_to_prover9(expression.second) + ')'
    else:
        return str(expression)


class Prover9(Prover9Parent, Prover):
    _prover9_bin = None
    _prooftrans_bin = None

    def __init__(self, timeout=60):
        self._timeout = timeout
        """The timeout value for prover9.  If a proof can not be found
           in this amount of time, then prover9 will return false.
           (Use 0 for no timeout.)"""

    def _prove(self, goal=None, assumptions=None, verbose=False):
        """
        Use Prover9 to prove a theorem.
        :return: A pair whose first element is a boolean indicating if the
        proof was successful (i.e. returns value of 0) and whose second element
        is the output of the prover.
        """
        if not assumptions:
            assumptions = []

        stdout, returncode = self._call_prover9(self.prover9_input(goal, assumptions),
                                                verbose=verbose)
        return (returncode == 0, stdout)

    def prover9_input(self, goal, assumptions):
        """
        :see: Prover9Parent.prover9_input
        """
        s = 'clear(auto_denials).\n' #only one proof required
        return s + Prover9Parent.prover9_input(self, goal, assumptions)

    def _call_prover9(self, input_str, args=[], verbose=False):
        """
        Call the ``prover9`` binary with the given input.

        :param input_str: A string whose contents are used as stdin.
        :param args: A list of command-line arguments.
        :return: A tuple (stdout, returncode)
        :see: ``config_prover9``
        """
        if self._prover9_bin is None:
            self._prover9_bin = self._find_binary('prover9', verbose)

        updated_input_str = ''
        if self._timeout > 0:
            updated_input_str += 'assign(max_seconds, %d).\n\n' % self._timeout
        updated_input_str += input_str

        stdout, returncode = self._call(updated_input_str, self._prover9_bin, args, verbose)

        if returncode not in [0,2]:
            errormsgprefix = '%%ERROR:'
            if errormsgprefix in stdout:
                msgstart = stdout.index(errormsgprefix)
                errormsg = stdout[msgstart:].strip()
            else:
                errormsg = None
            if returncode in [3,4,5,6]:
                raise Prover9LimitExceededException(returncode, errormsg)
            else:
                raise Prover9FatalException(returncode, errormsg)

        return stdout, returncode

    def _call_prooftrans(self, input_str, args=[], verbose=False):
        """
        Call the ``prooftrans`` binary with the given input.

        :param input_str: A string whose contents are used as stdin.
        :param args: A list of command-line arguments.
        :return: A tuple (stdout, returncode)
        :see: ``config_prover9``
        """
        if self._prooftrans_bin is None:
            self._prooftrans_bin = self._find_binary('prooftrans', verbose)

        return self._call(input_str, self._prooftrans_bin, args, verbose)


class Prover9Exception(Exception):
    def __init__(self, returncode, message):
        msg = p9_return_codes[returncode]
        if message:
            msg += '\n%s' % message
        Exception.__init__(self, msg)

class Prover9FatalException(Prover9Exception):
    pass

class Prover9LimitExceededException(Prover9Exception):
    pass



######################################################################
#{ Tests and Demos
######################################################################

def test_config():

    a = Expression.fromstring('(walk(j) & sing(j))')
    g = Expression.fromstring('walk(j)')
    p = Prover9Command(g, assumptions=[a])
    p._executable_path = None
    p.prover9_search=[]
    p.prove()
    #config_prover9('/usr/local/bin')
    print(p.prove())
    print(p.proof())

def test_convert_to_prover9(expr):
    """
    Test that parsing works OK.
    """
    for t in expr:
        e = Expression.fromstring(t)
        print(convert_to_prover9(e))

def test_prove(arguments):
    """
    Try some proofs and exhibit the results.
    """
    for (goal, assumptions) in arguments:
        g = Expression.fromstring(goal)
        alist = [Expression.fromstring(a) for a in assumptions]
        p = Prover9Command(g, assumptions=alist).prove()
        for a in alist:
            print('   %s' % a)
        print('|- %s: %s\n' % (g, p))

arguments = [
    ('(man(x) <-> (not (not man(x))))', []),
    ('(not (man(x) & (not man(x))))', []),
    ('(man(x) | (not man(x)))', []),
    ('(man(x) & (not man(x)))', []),
    ('(man(x) -> man(x))', []),
    ('(not (man(x) & (not man(x))))', []),
    ('(man(x) | (not man(x)))', []),
    ('(man(x) -> man(x))', []),
    ('(man(x) <-> man(x))', []),
    ('(not (man(x) <-> (not man(x))))', []),
    ('mortal(Socrates)', ['all x.(man(x) -> mortal(x))', 'man(Socrates)']),
    ('((all x.(man(x) -> walks(x)) & man(Socrates)) -> some y.walks(y))', []),
    ('(all x.man(x) -> all x.man(x))', []),
    ('some x.all y.sees(x,y)', []),
    ('some e3.(walk(e3) & subj(e3, mary))',
        ['some e1.(see(e1) & subj(e1, john) & some e2.(pred(e1, e2) & walk(e2) & subj(e2, mary)))']),
    ('some x e1.(see(e1) & subj(e1, x) & some e2.(pred(e1, e2) & walk(e2) & subj(e2, mary)))',
       ['some e1.(see(e1) & subj(e1, john) & some e2.(pred(e1, e2) & walk(e2) & subj(e2, mary)))'])
]

expressions = [r'some x y.sees(x,y)',
               r'some x.(man(x) & walks(x))',
               r'\x.(man(x) & walks(x))',
               r'\x y.sees(x,y)',
               r'walks(john)',
               r'\x.big(x, \y.mouse(y))',
               r'(walks(x) & (runs(x) & (threes(x) & fours(x))))',
               r'(walks(x) -> runs(x))',
               r'some x.(PRO(x) & sees(John, x))',
               r'some x.(man(x) & (not walks(x)))',
               r'all x.(man(x) -> walks(x))']

def spacer(num=45):
    print('-' * num)

def demo():
    print("Testing configuration")
    spacer()
    test_config()
    print()
    print("Testing conversion to Prover9 format")
    spacer()
    test_convert_to_prover9(expressions)
    print()
    print("Testing proofs")
    spacer()
    test_prove(arguments)

if __name__ == '__main__':
    demo()



# Natural Language Toolkit: Internal utility functions
#
# Copyright (C) 2001-2018 NLTK Project
# Author: Steven Bird <stevenbird1@gmail.com>
#         Edward Loper <edloper@gmail.com>
#         Nitin Madnani <nmadnani@ets.org>
# URL: <http://nltk.org/>
# For license information, see LICENSE.TXT
from __future__ import print_function

import subprocess
import os
import fnmatch
import re
import warnings
import textwrap
import types
import sys
import stat
import locale

# Use the c version of ElementTree, which is faster, if possible:
try:
    from xml.etree import cElementTree as ElementTree
except ImportError:
    from xml.etree import ElementTree

from six import string_types

from nltk import __file__
from nltk import compat

##########################################################################
# Java Via Command-Line
##########################################################################

_java_bin = None
_java_options = []
# [xx] add classpath option to config_java?
def config_java(bin=None, options=None, verbose=False):
    """
    Configure nltk's java interface, by letting nltk know where it can
    find the Java binary, and what extra options (if any) should be
    passed to Java when it is run.

    :param bin: The full path to the Java binary.  If not specified,
        then nltk will search the system for a Java binary; and if
        one is not found, it will raise a ``LookupError`` exception.
    :type bin: str
    :param options: A list of options that should be passed to the
        Java binary when it is called.  A common value is
        ``'-Xmx512m'``, which tells Java binary to increase
        the maximum heap size to 512 megabytes.  If no options are
        specified, then do not modify the options list.
    :type options: list(str)
    """
    global _java_bin, _java_options
    _java_bin = find_binary('java', bin, env_vars=['JAVAHOME', 'JAVA_HOME'], verbose=verbose, binary_names=['java.exe'])

    if options is not None:
        if isinstance(options, string_types):
            options = options.split()
        _java_options = list(options)

def java(cmd, classpath=None, stdin=None, stdout=None, stderr=None,
         blocking=True):
    """
    Execute the given java command, by opening a subprocess that calls
    Java.  If java has not yet been configured, it will be configured
    by calling ``config_java()`` with no arguments.

    :param cmd: The java command that should be called, formatted as
        a list of strings.  Typically, the first string will be the name
        of the java class; and the remaining strings will be arguments
        for that java class.
    :type cmd: list(str)

    :param classpath: A ``':'`` separated list of directories, JAR
        archives, and ZIP archives to search for class files.
    :type classpath: str

    :param stdin, stdout, stderr: Specify the executed programs'
        standard input, standard output and standard error file
        handles, respectively.  Valid values are ``subprocess.PIPE``,
        an existing file descriptor (a positive integer), an existing
        file object, and None.  ``subprocess.PIPE`` indicates that a
        new pipe to the child should be created.  With None, no
        redirection will occur; the child's file handles will be
        inherited from the parent.  Additionally, stderr can be
        ``subprocess.STDOUT``, which indicates that the stderr data
        from the applications should be captured into the same file
        handle as for stdout.

    :param blocking: If ``false``, then return immediately after
        spawning the subprocess.  In this case, the return value is
        the ``Popen`` object, and not a ``(stdout, stderr)`` tuple.

    :return: If ``blocking=True``, then return a tuple ``(stdout,
        stderr)``, containing the stdout and stderr outputs generated
        by the java command if the ``stdout`` and ``stderr`` parameters
        were set to ``subprocess.PIPE``; or None otherwise.  If
        ``blocking=False``, then return a ``subprocess.Popen`` object.

    :raise OSError: If the java command returns a nonzero return code.
    """
    if stdin == 'pipe': stdin = subprocess.PIPE
    if stdout == 'pipe': stdout = subprocess.PIPE
    if stderr == 'pipe': stderr = subprocess.PIPE
    if isinstance(cmd, string_types):
        raise TypeError('cmd should be a list of strings')

    # Make sure we know where a java binary is.
    if _java_bin is None:
        config_java()

    # Set up the classpath.
    if isinstance(classpath, string_types):
        classpaths=[classpath]
    else:
        classpaths=list(classpath)
    classpath=os.path.pathsep.join(classpaths)

    # Construct the full command string.
    cmd = list(cmd)
    cmd = ['-cp', classpath] + cmd
    cmd = [_java_bin] + _java_options + cmd

    # Call java via a subprocess
    p = subprocess.Popen(cmd, stdin=stdin, stdout=stdout, stderr=stderr)
    if not blocking: return p
    (stdout, stderr) = p.communicate()

    # Check the return code.
    if p.returncode != 0:
        print(_decode_stdoutdata(stderr))
        raise OSError('Java command failed : ' + str(cmd))

    return (stdout, stderr)

if 0:
    #config_java(options='-Xmx512m')
    # Write:
    #java('weka.classifiers.bayes.NaiveBayes',
    #     ['-d', '/tmp/names.model', '-t', '/tmp/train.arff'],
    #     classpath='/Users/edloper/Desktop/weka/weka.jar')
    # Read:
    (a,b) = java(['weka.classifiers.bayes.NaiveBayes',
                  '-l', '/tmp/names.model', '-T', '/tmp/test.arff',
                  '-p', '0'],#, '-distribution'],
                 classpath='/Users/edloper/Desktop/weka/weka.jar')


######################################################################
# Parsing
######################################################################

class ReadError(ValueError):
    """
    Exception raised by read_* functions when they fail.
    :param position: The index in the input string where an error occurred.
    :param expected: What was expected when an error occurred.
    """
    def __init__(self, expected, position):
        ValueError.__init__(self, expected, position)
        self.expected = expected
        self.position = position
    def __str__(self):
        return 'Expected %s at %s' % (self.expected, self.position)

_STRING_START_RE = re.compile(r"[uU]?[rR]?(\"\"\"|\'\'\'|\"|\')")
def read_str(s, start_position):
    """
    If a Python string literal begins at the specified position in the
    given string, then return a tuple ``(val, end_position)``
    containing the value of the string literal and the position where
    it ends.  Otherwise, raise a ``ReadError``.

    :param s: A string that will be checked to see if within which a
        Python string literal exists.
    :type s: str

    :param start_position: The specified beginning position of the string ``s``
        to begin regex matching.
    :type start_position: int

    :return: A tuple containing the matched string literal evaluated as a
        string and the end position of the string literal.
    :rtype: tuple(str, int)

    :raise ReadError: If the ``_STRING_START_RE`` regex doesn't return a
        match in ``s`` at ``start_position``, i.e., open quote. If the
        ``_STRING_END_RE`` regex doesn't return a match in ``s`` at the
        end of the first match, i.e., close quote.
    :raise ValueError: If an invalid string (i.e., contains an invalid
        escape sequence) is passed into the ``eval``.

    :Example:
    >>> from nltk.internals import read_str
    >>> read_str('"Hello", World!', 0)
    ('Hello', 7)

    """
    # Read the open quote, and any modifiers.
    m = _STRING_START_RE.match(s, start_position)
    if not m: raise ReadError('open quote', start_position)
    quotemark = m.group(1)

    # Find the close quote.
    _STRING_END_RE = re.compile(r'\\|%s' % quotemark)
    position = m.end()
    while True:
        match = _STRING_END_RE.search(s, position)
        if not match: raise ReadError('close quote', position)
        if match.group(0) == '\\': position = match.end()+1
        else: break

    # Process it, using eval.  Strings with invalid escape sequences
    # might raise ValueEerror.
    try:
        return eval(s[start_position:match.end()]), match.end()
    except ValueError as e:
        raise ReadError('invalid string (%s)' % e)

_READ_INT_RE = re.compile(r'-?\d+')
def read_int(s, start_position):
    """
    If an integer begins at the specified position in the given
    string, then return a tuple ``(val, end_position)`` containing the
    value of the integer and the position where it ends.  Otherwise,
    raise a ``ReadError``.

    :param s: A string that will be checked to see if within which a
        Python integer exists.
    :type s: str

    :param start_position: The specified beginning position of the string ``s``
        to begin regex matching.
    :type start_position: int

    :return: A tuple containing the matched integer casted to an int,
        and the end position of the int in ``s``.
    :rtype: tuple(int, int)

    :raise ReadError: If the ``_READ_INT_RE`` regex doesn't return a
        match in ``s`` at ``start_position``.

    :Example:
    >>> from nltk.internals import read_int
    >>> read_int('42 is the answer', 0)
    (42, 2)

    """
    m = _READ_INT_RE.match(s, start_position)
    if not m: raise ReadError('integer', start_position)
    return int(m.group()), m.end()

_READ_NUMBER_VALUE = re.compile(r'-?(\d*)([.]?\d*)?')
def read_number(s, start_position):
    """
    If an integer or float begins at the specified position in the
    given string, then return a tuple ``(val, end_position)``
    containing the value of the number and the position where it ends.
    Otherwise, raise a ``ReadError``.

    :param s: A string that will be checked to see if within which a
        Python number exists.
    :type s: str

    :param start_position: The specified beginning position of the string ``s``
        to begin regex matching.
    :type start_position: int

    :return: A tuple containing the matched number casted to a ``float``,
        and the end position of the number in ``s``.
    :rtype: tuple(float, int)

    :raise ReadError: If the ``_READ_NUMBER_VALUE`` regex doesn't return a
        match in ``s`` at ``start_position``.

    :Example:
    >>> from nltk.internals import read_number
    >>> read_number('Pi is 3.14159', 6)
    (3.14159, 13)

    """
    m = _READ_NUMBER_VALUE.match(s, start_position)
    if not m or not (m.group(1) or m.group(2)):
        raise ReadError('number', start_position)
    if m.group(2): return float(m.group()), m.end()
    else: return int(m.group()), m.end()



######################################################################
# Check if a method has been overridden
######################################################################

def overridden(method):
    """
    :return: True if ``method`` overrides some method with the same
    name in a base class.  This is typically used when defining
    abstract base classes or interfaces, to allow subclasses to define
    either of two related methods:

        >>> class EaterI:
        ...     '''Subclass must define eat() or batch_eat().'''
        ...     def eat(self, food):
        ...         if overridden(self.batch_eat):
        ...             return self.batch_eat([food])[0]
        ...         else:
        ...             raise NotImplementedError()
        ...     def batch_eat(self, foods):
        ...         return [self.eat(food) for food in foods]

    :type method: instance method
    """
    # [xx] breaks on classic classes!
    if isinstance(method, types.MethodType) and compat.get_im_class(method) is not None:
        name = method.__name__
        funcs = [cls.__dict__[name]
                 for cls in _mro(compat.get_im_class(method))
                 if name in cls.__dict__]
        return len(funcs) > 1
    else:
        raise TypeError('Expected an instance method.')

def _mro(cls):
    """
    Return the method resolution order for ``cls`` -- i.e., a list
    containing ``cls`` and all its base classes, in the order in which
    they would be checked by ``getattr``.  For new-style classes, this
    is just cls.__mro__.  For classic classes, this can be obtained by
    a depth-first left-to-right traversal of ``__bases__``.
    """
    if isinstance(cls, type):
        return cls.__mro__
    else:
        mro = [cls]
        for base in cls.__bases__: mro.extend(_mro(base))
        return mro

######################################################################
# Deprecation decorator & base class
######################################################################
# [xx] dedent msg first if it comes from  a docstring.

def _add_epytext_field(obj, field, message):
    """Add an epytext @field to a given object's docstring."""
    indent = ''
    # If we already have a docstring, then add a blank line to separate
    # it from the new field, and check its indentation.
    if obj.__doc__:
        obj.__doc__ = obj.__doc__.rstrip()+'\n\n'
        indents = re.findall(r'(?<=\n)[ ]+(?!\s)', obj.__doc__.expandtabs())
        if indents: indent = min(indents)
    # If we don't have a docstring, add an empty one.
    else:
        obj.__doc__ = ''

    obj.__doc__ += textwrap.fill('@%s: %s' % (field, message),
                                 initial_indent=indent,
                                 subsequent_indent=indent+'    ')

def deprecated(message):
    """
    A decorator used to mark functions as deprecated.  This will cause
    a warning to be printed the when the function is used.  Usage:

        >>> from nltk.internals import deprecated
        >>> @deprecated('Use foo() instead')
        ... def bar(x):
        ...     print(x/10)

    """

    def decorator(func):
        msg = ("Function %s() has been deprecated.  %s"
               % (func.__name__, message))
        msg = '\n' + textwrap.fill(msg, initial_indent='  ',
                                   subsequent_indent='  ')
        def newFunc(*args, **kwargs):
            warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        # Copy the old function's name, docstring, & dict
        newFunc.__dict__.update(func.__dict__)
        newFunc.__name__ = func.__name__
        newFunc.__doc__ = func.__doc__
        newFunc.__deprecated__ = True
        # Add a @deprecated field to the docstring.
        _add_epytext_field(newFunc, 'deprecated', message)
        return newFunc
    return decorator

class Deprecated(object):
    """
    A base class used to mark deprecated classes.  A typical usage is to
    alert users that the name of a class has changed:

        >>> from nltk.internals import Deprecated
        >>> class NewClassName(object):
        ...     pass # All logic goes here.
        ...
        >>> class OldClassName(Deprecated, NewClassName):
        ...     "Use NewClassName instead."

    The docstring of the deprecated class will be used in the
    deprecation warning message.
    """
    def __new__(cls, *args, **kwargs):
        # Figure out which class is the deprecated one.
        dep_cls = None
        for base in _mro(cls):
            if Deprecated in base.__bases__:
                dep_cls = base; break
        assert dep_cls, 'Unable to determine which base is deprecated.'

        # Construct an appropriate warning.
        doc = dep_cls.__doc__ or ''.strip()
        # If there's a @deprecated field, strip off the field marker.
        doc = re.sub(r'\A\s*@deprecated:', r'', doc)
        # Strip off any indentation.
        doc = re.sub(r'(?m)^\s*', '', doc)
        # Construct a 'name' string.
        name = 'Class %s' % dep_cls.__name__
        if cls != dep_cls:
            name += ' (base class for %s)' % cls.__name__
        # Put it all together.
        msg = '%s has been deprecated.  %s' % (name, doc)
        # Wrap it.
        msg = '\n' + textwrap.fill(msg, initial_indent='    ',
                                   subsequent_indent='    ')
        warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
        # Do the actual work of __new__.
        return object.__new__(cls)

##########################################################################
# COUNTER, FOR UNIQUE NAMING
##########################################################################

class Counter:
    """
    A counter that auto-increments each time its value is read.
    """
    def __init__(self, initial_value=0):
        self._value = initial_value
    def get(self):
        self._value += 1
        return self._value

##########################################################################
# Search for files/binaries
##########################################################################

def find_file_iter(filename, env_vars=(), searchpath=(),
    file_names=None, url=None, verbose=False, finding_dir=False):
    """
    Search for a file to be used by nltk.

    :param filename: The name or path of the file.
    :param env_vars: A list of environment variable names to check.
    :param file_names: A list of alternative file names to check.
    :param searchpath: List of directories to search.
    :param url: URL presented to user for download help.
    :param verbose: Whether or not to print path when a file is found.
    """
    file_names = [filename] + (file_names or [])
    assert isinstance(filename, string_types)
    assert not isinstance(file_names, string_types)
    assert not isinstance(searchpath, string_types)
    if isinstance(env_vars, string_types):
        env_vars = env_vars.split()
    yielded = False

    # File exists, no magic
    for alternative in file_names:
        path_to_file = os.path.join(filename, alternative)
        if os.path.isfile(path_to_file):
            if verbose:
                print('[Found %s: %s]' % (filename, path_to_file))
            yielded = True
            yield path_to_file
        # Check the bare alternatives
        if os.path.isfile(alternative):
            if verbose:
                print('[Found %s: %s]' % (filename, alternative))
            yielded = True
            yield alternative
        # Check if the alternative is inside a 'file' directory
        path_to_file = os.path.join(filename, 'file', alternative)
        if os.path.isfile(path_to_file):
            if verbose:
                print('[Found %s: %s]' % (filename, path_to_file))
            yielded = True
            yield path_to_file

    # Check environment variables
    for env_var in env_vars:
        if env_var in os.environ:
            if finding_dir: # This is to file a directory instead of file
                yielded = True
                yield os.environ[env_var]

            for env_dir in os.environ[env_var].split(os.pathsep):
                # Check if the environment variable contains a direct path to the bin
                if os.path.isfile(env_dir):
                    if verbose:
                        print('[Found %s: %s]'%(filename, env_dir))
                    yielded = True
                    yield env_dir
                # Check if the possible bin names exist inside the environment variable directories
                for alternative in file_names:
                    path_to_file = os.path.join(env_dir, alternative)
                    if os.path.isfile(path_to_file):
                        if verbose:
                            print('[Found %s: %s]'%(filename, path_to_file))
                        yielded = True
                        yield path_to_file
                    # Check if the alternative is inside a 'file' directory
                    # path_to_file = os.path.join(env_dir, 'file', alternative)

                    # Check if the alternative is inside a 'bin' directory
                    path_to_file = os.path.join(env_dir, 'bin', alternative)

                    if os.path.isfile(path_to_file):
                        if verbose:
                            print('[Found %s: %s]' % (filename, path_to_file))
                        yielded = True
                        yield path_to_file

    # Check the path list.
    for directory in searchpath:
        for alternative in file_names:
            path_to_file = os.path.join(directory, alternative)
            if os.path.isfile(path_to_file):
                yielded = True
                yield path_to_file

    # If we're on a POSIX system, then try using the 'which' command
    # to find the file.
    if os.name == 'posix':
        for alternative in file_names:
            try:
                p = subprocess.Popen(['which', alternative],
                        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                stdout, stderr = p.communicate()
                path = _decode_stdoutdata(stdout).strip()
                if path.endswith(alternative) and os.path.exists(path):
                    if verbose:
                        print('[Found %s: %s]' % (filename, path))
                    yielded = True
                    yield path
            except (KeyboardInterrupt, SystemExit, OSError):
                raise
            except:
                pass

    if not yielded:
        msg = ("NLTK was unable to find the %s file!" "\nUse software specific "
               "configuration paramaters" % filename)
        if env_vars: msg += ' or set the %s environment variable' % env_vars[0]
        msg += '.'
        if searchpath:
            msg += '\n\n  Searched in:'
            msg += ''.join('\n    - %s' % d for d in searchpath)
        if url: msg += ('\n\n  For more information on %s, see:\n    <%s>' %
                        (filename, url))
        div = '='*75
        raise LookupError('\n\n%s\n%s\n%s' % (div, msg, div))


def find_file(filename, env_vars=(), searchpath=(),
        file_names=None, url=None, verbose=False):
    return next(find_file_iter(filename, env_vars, searchpath,
                               file_names, url, verbose))


def find_dir(filename, env_vars=(), searchpath=(),
        file_names=None, url=None, verbose=False):
    return next(find_file_iter(filename, env_vars, searchpath,
                               file_names, url, verbose, finding_dir=True))


def find_binary_iter(name, path_to_bin=None, env_vars=(), searchpath=(),
                binary_names=None, url=None, verbose=False):
    """
    Search for a file to be used by nltk.

    :param name: The name or path of the file.
    :param path_to_bin: The user-supplied binary location (deprecated)
    :param env_vars: A list of environment variable names to check.
    :param file_names: A list of alternative file names to check.
    :param searchpath: List of directories to search.
    :param url: URL presented to user for download help.
    :param verbose: Whether or not to print path when a file is found.
    """
    for file in  find_file_iter(path_to_bin or name, env_vars, searchpath, binary_names,
                     url, verbose):
        yield file

def find_binary(name, path_to_bin=None, env_vars=(), searchpath=(),
                binary_names=None, url=None, verbose=False):
    return next(find_binary_iter(name, path_to_bin, env_vars, searchpath,
                                 binary_names, url, verbose))

def find_jar_iter(name_pattern, path_to_jar=None, env_vars=(),
        searchpath=(), url=None, verbose=False, is_regex=False):
    """
    Search for a jar that is used by nltk.

    :param name_pattern: The name of the jar file
    :param path_to_jar: The user-supplied jar location, or None.
    :param env_vars: A list of environment variable names to check
                     in addition to the CLASSPATH variable which is
                     checked by default.
    :param searchpath: List of directories to search.
    :param is_regex: Whether name is a regular expression.
    """

    assert isinstance(name_pattern, string_types)
    assert not isinstance(searchpath, string_types)
    if isinstance(env_vars, string_types):
        env_vars = env_vars.split()
    yielded = False

    # Make sure we check the CLASSPATH first
    env_vars = ['CLASSPATH'] + list(env_vars)

    # If an explicit location was given, then check it, and yield it if
    # it's present; otherwise, complain.
    if path_to_jar is not None:
        if os.path.isfile(path_to_jar):
            yielded = True
            yield path_to_jar
        else:
            raise LookupError('Could not find %s jar file at %s' %
                            (name_pattern, path_to_jar))

    # Check environment variables
    for env_var in env_vars:
        if env_var in os.environ:
            if env_var == 'CLASSPATH':
                classpath = os.environ['CLASSPATH']
                for cp in classpath.split(os.path.pathsep):
                    if os.path.isfile(cp):
                        filename=os.path.basename(cp)
                        if is_regex and re.match(name_pattern, filename) or \
                                (not is_regex and filename == name_pattern):
                            if verbose:
                                print('[Found %s: %s]' % (name_pattern, cp))
                            yielded = True
                            yield cp
                    # The case where user put directory containing the jar file in the classpath
                    if os.path.isdir(cp):
                        if not is_regex:
                            if os.path.isfile(os.path.join(cp,name_pattern)):
                                if verbose:
                                    print('[Found %s: %s]' % (name_pattern, cp))
                                yielded = True
                                yield os.path.join(cp,name_pattern)
                        else:
                            # Look for file using regular expression
                            for file_name in os.listdir(cp):
                                if re.match(name_pattern,file_name):
                                    if verbose:
                                        print('[Found %s: %s]' % (name_pattern, os.path.join(cp,file_name)))
                                    yielded = True
                                    yield os.path.join(cp,file_name)

            else:
                jar_env = os.environ[env_var]
                jar_iter = ((os.path.join(jar_env, path_to_jar) for path_to_jar in os.listdir(jar_env))
                            if os.path.isdir(jar_env) else (jar_env,))
                for path_to_jar in jar_iter:
                    if os.path.isfile(path_to_jar):
                        filename=os.path.basename(path_to_jar)
                        if is_regex and re.match(name_pattern, filename) or \
                                (not is_regex and filename == name_pattern):
                            if verbose:
                                print('[Found %s: %s]' % (name_pattern, path_to_jar))
                            yielded = True
                            yield path_to_jar

    # Check the path list.
    for directory in searchpath:
        if is_regex:
            for filename in os.listdir(directory):
                path_to_jar = os.path.join(directory, filename)
                if os.path.isfile(path_to_jar):
                    if re.match(name_pattern, filename):
                        if verbose:
                            print('[Found %s: %s]' % (filename, path_to_jar))
                yielded = True
                yield path_to_jar
        else:
            path_to_jar = os.path.join(directory, name_pattern)
            if os.path.isfile(path_to_jar):
                if verbose:
                    print('[Found %s: %s]' % (name_pattern, path_to_jar))
                yielded = True
                yield path_to_jar

    if not yielded:
        # If nothing was found, raise an error
        msg = ("NLTK was unable to find %s!" % name_pattern)
        if env_vars: msg += ' Set the %s environment variable' % env_vars[0]
        msg = textwrap.fill(msg+'.', initial_indent='  ',
                            subsequent_indent='  ')
        if searchpath:
            msg += '\n\n  Searched in:'
            msg += ''.join('\n    - %s' % d for d in searchpath)
        if url:
            msg += ('\n\n  For more information, on %s, see:\n    <%s>' %
                    (name_pattern, url))
        div = '='*75
        raise LookupError('\n\n%s\n%s\n%s' % (div, msg, div))

def find_jar(name_pattern, path_to_jar=None, env_vars=(),
        searchpath=(), url=None, verbose=False, is_regex=False):
    return next(find_jar_iter(name_pattern, path_to_jar, env_vars,
                         searchpath, url, verbose, is_regex))


def find_jars_within_path(path_to_jars):
	return [os.path.join(root, filename)
			for root, dirnames, filenames in os.walk(path_to_jars)
			for filename in fnmatch.filter(filenames, '*.jar')]

def _decode_stdoutdata(stdoutdata):
    """ Convert data read from stdout/stderr to unicode """
    if not isinstance(stdoutdata, bytes):
        return stdoutdata

    encoding = getattr(sys.__stdout__, "encoding", locale.getpreferredencoding())
    if encoding is None:
        return stdoutdata.decode()
    return stdoutdata.decode(encoding)

##########################################################################
# Import Stdlib Module
##########################################################################

def import_from_stdlib(module):
    """
    When python is run from within the nltk/ directory tree, the
    current directory is included at the beginning of the search path.
    Unfortunately, that means that modules within nltk can sometimes
    shadow standard library modules.  As an example, the stdlib
    'inspect' module will attempt to import the stdlib 'tokenize'
    module, but will instead end up importing NLTK's 'tokenize' module
    instead (causing the import to fail).
    """
    old_path = sys.path
    sys.path = [d for d in sys.path if d not in ('', '.')]
    m = __import__(module)
    sys.path = old_path
    return m


##########################################################################
# Wrapper for ElementTree Elements
##########################################################################

@compat.python_2_unicode_compatible
class ElementWrapper(object):
    """
    A wrapper around ElementTree Element objects whose main purpose is
    to provide nicer __repr__ and __str__ methods.  In addition, any
    of the wrapped Element's methods that return other Element objects
    are overridden to wrap those values before returning them.

    This makes Elements more convenient to work with in
    interactive sessions and doctests, at the expense of some
    efficiency.
    """

    # Prevent double-wrapping:
    def __new__(cls, etree):
        """
        Create and return a wrapper around a given Element object.
        If ``etree`` is an ``ElementWrapper``, then ``etree`` is
        returned as-is.
        """
        if isinstance(etree, ElementWrapper):
            return etree
        else:
            return object.__new__(ElementWrapper)

    def __init__(self, etree):
        r"""
        Initialize a new Element wrapper for ``etree``.

        If ``etree`` is a string, then it will be converted to an
        Element object using ``ElementTree.fromstring()`` first:

            >>> ElementWrapper("<test></test>")
            <Element "<?xml version='1.0' encoding='utf8'?>\n<test />">

        """
        if isinstance(etree, string_types):
            etree = ElementTree.fromstring(etree)
        self.__dict__['_etree'] = etree

    def unwrap(self):
        """
        Return the Element object wrapped by this wrapper.
        """
        return self._etree

    ##////////////////////////////////////////////////////////////
    #{ String Representation
    ##////////////////////////////////////////////////////////////

    def __repr__(self):
        s = ElementTree.tostring(self._etree, encoding='utf8').decode('utf8')
        if len(s) > 60:
            e = s.rfind('<')
            if (len(s)-e) > 30: e = -20
            s = '%s...%s' % (s[:30], s[e:])
        return '<Element %r>' % s

    def __str__(self):
        """
        :return: the result of applying ``ElementTree.tostring()`` to
        the wrapped Element object.
        """
        return ElementTree.tostring(self._etree, encoding='utf8').decode('utf8').rstrip()

    ##////////////////////////////////////////////////////////////
    #{ Element interface Delegation (pass-through)
    ##////////////////////////////////////////////////////////////

    def __getattr__(self, attrib):
        return getattr(self._etree, attrib)

    def __setattr__(self, attr, value):
        return setattr(self._etree, attr, value)

    def __delattr__(self, attr):
        return delattr(self._etree, attr)

    def __setitem__(self, index, element):
        self._etree[index] = element

    def __delitem__(self, index):
        del self._etree[index]

    def __setslice__(self, start, stop, elements):
        self._etree[start:stop] = elements

    def __delslice__(self, start, stop):
        del self._etree[start:stop]

    def __len__(self):
        return len(self._etree)

    ##////////////////////////////////////////////////////////////
    #{ Element interface Delegation (wrap result)
    ##////////////////////////////////////////////////////////////

    def __getitem__(self, index):
        return ElementWrapper(self._etree[index])

    def __getslice__(self, start, stop):
        return [ElementWrapper(elt) for elt in self._etree[start:stop]]

    def getchildren(self):
        return [ElementWrapper(elt) for elt in self._etree]

    def getiterator(self, tag=None):
        return (ElementWrapper(elt)
                for elt in self._etree.getiterator(tag))

    def makeelement(self, tag, attrib):
        return ElementWrapper(self._etree.makeelement(tag, attrib))

    def find(self, path):
        elt = self._etree.find(path)
        if elt is None: return elt
        else: return ElementWrapper(elt)

    def findall(self, path):
        return [ElementWrapper(elt) for elt in self._etree.findall(path)]

######################################################################
# Helper for Handling Slicing
######################################################################

def slice_bounds(sequence, slice_obj, allow_step=False):
    """
    Given a slice, return the corresponding (start, stop) bounds,
    taking into account None indices and negative indices.  The
    following guarantees are made for the returned start and stop values:

      - 0 <= start <= len(sequence)
      - 0 <= stop <= len(sequence)
      - start <= stop

    :raise ValueError: If ``slice_obj.step`` is not None.
    :param allow_step: If true, then the slice object may have a
        non-None step.  If it does, then return a tuple
        (start, stop, step).
    """
    start, stop = (slice_obj.start, slice_obj.stop)

    # If allow_step is true, then include the step in our return
    # value tuple.
    if allow_step:
        step = slice_obj.step
        if step is None: step = 1
        # Use a recursive call without allow_step to find the slice
        # bounds.  If step is negative, then the roles of start and
        # stop (in terms of default values, etc), are swapped.
        if step < 0:
            start, stop = slice_bounds(sequence, slice(stop, start))
        else:
            start, stop = slice_bounds(sequence, slice(start, stop))
        return start, stop, step

    # Otherwise, make sure that no non-default step value is used.
    elif slice_obj.step not in (None, 1):
        raise ValueError('slices with steps are not supported by %s' %
                         sequence.__class__.__name__)

    # Supply default offsets.
    if start is None: start = 0
    if stop is None: stop = len(sequence)

    # Handle negative indices.
    if start < 0: start = max(0, len(sequence)+start)
    if stop < 0: stop = max(0, len(sequence)+stop)

    # Make sure stop doesn't go past the end of the list.  Note that
    # we avoid calculating len(sequence) if possible, because for lazy
    # sequences, calculating the length of a sequence can be expensive.
    if stop > 0:
        try: sequence[stop-1]
        except IndexError: stop = len(sequence)

    # Make sure start isn't past stop.
    start = min(start, stop)

    # That's all folks!
    return start, stop

######################################################################
# Permission Checking
######################################################################

def is_writable(path):
    # Ensure that it exists.
    if not os.path.exists(path):
        return False

    # If we're on a posix system, check its permissions.
    if hasattr(os, 'getuid'):
        statdata = os.stat(path)
        perm = stat.S_IMODE(statdata.st_mode)
        # is it world-writable?
        if (perm & 0o002):
            return True
        # do we own it?
        elif statdata.st_uid == os.getuid() and (perm & 0o200):
            return True
        # are we in a group that can write to it?
        elif (statdata.st_gid in [os.getgid()] + os.getgroups()) \
            and (perm & 0o020):
            return True
        # otherwise, we can't write to it.
        else:
            return False

    # Otherwise, we'll assume it's writable.
    # [xx] should we do other checks on other platforms?
    return True

######################################################################
# NLTK Error reporting
######################################################################

def raise_unorderable_types(ordering, a, b):
    raise TypeError("unorderable types: %s() %s %s()" % (type(a).__name__, ordering, type(b).__name__))



# Natural Language Toolkit: Naive Bayes Classifiers
#
# Copyright (C) 2001-2018 NLTK Project
# Author: Edward Loper <edloper@gmail.com>
# URL: <http://nltk.org/>
# For license information, see LICENSE.TXT

"""
A classifier based on the Naive Bayes algorithm.  In order to find the
probability for a label, this algorithm first uses the Bayes rule to
express P(label|features) in terms of P(label) and P(features|label):

|                       P(label) * P(features|label)
|  P(label|features) = ------------------------------
|                              P(features)

The algorithm then makes the 'naive' assumption that all features are
independent, given the label:

|                       P(label) * P(f1|label) * ... * P(fn|label)
|  P(label|features) = --------------------------------------------
|                                         P(features)

Rather than computing P(features) explicitly, the algorithm just
calculates the numerator for each label, and normalizes them so they
sum to one:

|                       P(label) * P(f1|label) * ... * P(fn|label)
|  P(label|features) = --------------------------------------------
|                        SUM[l]( P(l) * P(f1|l) * ... * P(fn|l) )
"""
from __future__ import print_function, unicode_literals

from collections import defaultdict

from nltk.probability import FreqDist, DictionaryProbDist, ELEProbDist, sum_logs
from nltk.classify.api import ClassifierI

##//////////////////////////////////////////////////////
##  Naive Bayes Classifier
##//////////////////////////////////////////////////////

class NaiveBayesClassifier(ClassifierI):
    """
    A Naive Bayes classifier.  Naive Bayes classifiers are
    paramaterized by two probability distributions:

      - P(label) gives the probability that an input will receive each
        label, given no information about the input's features.

      - P(fname=fval|label) gives the probability that a given feature
        (fname) will receive a given value (fval), given that the
        label (label).

    If the classifier encounters an input with a feature that has
    never been seen with any label, then rather than assigning a
    probability of 0 to all labels, it will ignore that feature.

    The feature value 'None' is reserved for unseen feature values;
    you generally should not use 'None' as a feature value for one of
    your own features.
    """
    def __init__(self, label_probdist, feature_probdist):
        """
        :param label_probdist: P(label), the probability distribution
            over labels.  It is expressed as a ``ProbDistI`` whose
            samples are labels.  I.e., P(label) =
            ``label_probdist.prob(label)``.

        :param feature_probdist: P(fname=fval|label), the probability
            distribution for feature values, given labels.  It is
            expressed as a dictionary whose keys are ``(label, fname)``
            pairs and whose values are ``ProbDistI`` objects over feature
            values.  I.e., P(fname=fval|label) =
            ``feature_probdist[label,fname].prob(fval)``.  If a given
            ``(label,fname)`` is not a key in ``feature_probdist``, then
            it is assumed that the corresponding P(fname=fval|label)
            is 0 for all values of ``fval``.
        """
        self._label_probdist = label_probdist
        self._feature_probdist = feature_probdist
        self._labels = list(label_probdist.samples())

    def labels(self):
        return self._labels

    def classify(self, featureset):
        return self.prob_classify(featureset).max()

    def prob_classify(self, featureset):
        # Discard any feature names that we've never seen before.
        # Otherwise, we'll just assign a probability of 0 to
        # everything.
        featureset = featureset.copy()
        for fname in list(featureset.keys()):
            for label in self._labels:
                if (label, fname) in self._feature_probdist:
                    break
            else:
                #print 'Ignoring unseen feature %s' % fname
                del featureset[fname]

        # Find the log probabilty of each label, given the features.
        # Start with the log probability of the label itself.
        logprob = {}
        for label in self._labels:
            logprob[label] = self._label_probdist.logprob(label)

        # Then add in the log probability of features given labels.
        for label in self._labels:
            for (fname, fval) in featureset.items():
                if (label, fname) in self._feature_probdist:
                    feature_probs = self._feature_probdist[label, fname]
                    logprob[label] += feature_probs.logprob(fval)
                else:
                    # nb: This case will never come up if the
                    # classifier was created by
                    # NaiveBayesClassifier.train().
                    logprob[label] += sum_logs([]) # = -INF.

        return DictionaryProbDist(logprob, normalize=True, log=True)

    def show_most_informative_features(self, n=10):
        # Determine the most relevant features, and display them.
        cpdist = self._feature_probdist
        print('Most Informative Features')

        for (fname, fval) in self.most_informative_features(n):
            def labelprob(l):
                return cpdist[l, fname].prob(fval)

            labels = sorted([l for l in self._labels
                             if fval in cpdist[l, fname].samples()],
                            key=labelprob)
            if len(labels) == 1:
                continue
            l0 = labels[0]
            l1 = labels[-1]
            if cpdist[l0, fname].prob(fval) == 0:
                ratio = 'INF'
            else:
                ratio = '%8.1f' % (cpdist[l1, fname].prob(fval) /
                                   cpdist[l0, fname].prob(fval))
            print(('%24s = %-14r %6s : %-6s = %s : 1.0' %
                   (fname, fval, ("%s" % l1)[:6], ("%s" % l0)[:6], ratio)))

    def most_informative_features(self, n=100):
        """
        Return a list of the 'most informative' features used by this
        classifier.  For the purpose of this function, the
        informativeness of a feature ``(fname,fval)`` is equal to the
        highest value of P(fname=fval|label), for any label, divided by
        the lowest value of P(fname=fval|label), for any label:

        |  max[ P(fname=fval|label1) / P(fname=fval|label2) ]
        """
        if hasattr(self, '_most_informative_features'):
            return self._most_informative_features[:n]
        else:
            # The set of (fname, fval) pairs used by this classifier.
            features = set()
            # The max & min probability associated w/ each (fname, fval)
            # pair.  Maps (fname,fval) -> float.
            maxprob = defaultdict(lambda: 0.0)
            minprob = defaultdict(lambda: 1.0)

            for (label, fname), probdist in self._feature_probdist.items():
                for fval in probdist.samples():
                    feature = (fname, fval)
                    features.add(feature)
                    p = probdist.prob(fval)
                    maxprob[feature] = max(p, maxprob[feature])
                    minprob[feature] = min(p, minprob[feature])
                    if minprob[feature] == 0:
                        features.discard(feature)

            # Convert features to a list, & sort it by how informative
            # features are.
            self._most_informative_features = sorted(features,
                                                key=lambda feature_:
                                                minprob[feature_]/maxprob[feature_])
        return self._most_informative_features[:n]

    @classmethod
    def train(cls, labeled_featuresets, estimator=ELEProbDist):
        """
        :param labeled_featuresets: A list of classified featuresets,
            i.e., a list of tuples ``(featureset, label)``.
        """
        label_freqdist = FreqDist()
        feature_freqdist = defaultdict(FreqDist)
        feature_values = defaultdict(set)
        fnames = set()

        # Count up how many times each feature value occurred, given
        # the label and featurename.
        for featureset, label in labeled_featuresets:
            label_freqdist[label] += 1
            for fname, fval in featureset.items():
                # Increment freq(fval|label, fname)
                feature_freqdist[label, fname][fval] += 1
                # Record that fname can take the value fval.
                feature_values[fname].add(fval)
                # Keep a list of all feature names.
                fnames.add(fname)

        # If a feature didn't have a value given for an instance, then
        # we assume that it gets the implicit value 'None.'  This loop
        # counts up the number of 'missing' feature values for each
        # (label,fname) pair, and increments the count of the fval
        # 'None' by that amount.
        for label in label_freqdist:
            num_samples = label_freqdist[label]
            for fname in fnames:
                count = feature_freqdist[label, fname].N()
                # Only add a None key when necessary, i.e. if there are
                # any samples with feature 'fname' missing.
                if num_samples - count > 0:
                    feature_freqdist[label, fname][None] += num_samples - count
                    feature_values[fname].add(None)

        # Create the P(label) distribution
        label_probdist = estimator(label_freqdist)

        # Create the P(fval|label, fname) distribution
        feature_probdist = {}
        for ((label, fname), freqdist) in feature_freqdist.items():
            probdist = estimator(freqdist, bins=len(feature_values[fname]))
            feature_probdist[label, fname] = probdist

        return cls(label_probdist, feature_probdist)

##//////////////////////////////////////////////////////
##  Demo
##//////////////////////////////////////////////////////

def demo():
    from nltk.classify.util import names_demo
    classifier = names_demo(NaiveBayesClassifier.train)
    classifier.show_most_informative_features()

if __name__ == '__main__':
    demo()


